"""
Optimization rules for Mock-Spark query optimizer.

Provides specific optimization rules for different query patterns.
"""

from typing import List, Set, cast
from .query_optimizer import OptimizationRule, Operation, OperationType


class FilterPushdownRule(OptimizationRule):
    """Push filters as early as possible in the query plan."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Move filter operations before other operations when possible."""
        if not self.can_apply(operations):
            return operations

        optimized = []
        filters = []

        for op in operations:
            if op.type == OperationType.FILTER:
                filters.extend(op.predicates)
            else:
                # If we have filters and this is a non-filter operation,
                # try to push filters before it
                if filters and self._can_push_filters_before(op):
                    # Create filter operation
                    filter_op = Operation(
                        type=OperationType.FILTER,
                        columns=[],
                        predicates=filters.copy(),
                        join_conditions=[],
                        group_by_columns=[],
                        order_by_columns=[],
                        limit_count=None,
                        window_specs=[],
                        metadata={},
                    )
                    optimized.append(filter_op)
                    filters = []

                optimized.append(op)

        # Add any remaining filters at the end
        if filters:
            filter_op = Operation(
                type=OperationType.FILTER,
                columns=[],
                predicates=filters,
                join_conditions=[],
                group_by_columns=[],
                order_by_columns=[],
                limit_count=None,
                window_specs=[],
                metadata={},
            )
            optimized.append(filter_op)

        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if filter pushdown can be applied."""
        filter_ops = [op for op in operations if op.type == OperationType.FILTER]
        return len(filter_ops) > 0

    def _can_push_filters_before(self, op: Operation) -> bool:
        """Check if filters can be pushed before this operation."""
        # Can't push filters before GROUP_BY, ORDER_BY, or WINDOW
        return op.type not in [
            OperationType.GROUP_BY,
            OperationType.ORDER_BY,
            OperationType.WINDOW,
        ]


class ColumnPruningRule(OptimizationRule):
    """Remove unnecessary columns from the query plan."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Remove columns that are not needed in the final result."""
        if not self.can_apply(operations):
            return operations

        # Find all columns that are actually needed
        needed_columns = self._find_needed_columns(operations)

        optimized = []
        for op in operations:
            if op.type == OperationType.SELECT:
                # Only select needed columns
                needed_for_op = [col for col in op.columns if col in needed_columns]
                if needed_for_op:
                    optimized_op = Operation(
                        type=op.type,
                        columns=needed_for_op,
                        predicates=op.predicates,
                        join_conditions=op.join_conditions,
                        group_by_columns=op.group_by_columns,
                        order_by_columns=op.order_by_columns,
                        limit_count=op.limit_count,
                        window_specs=op.window_specs,
                        metadata=op.metadata,
                    )
                    optimized.append(optimized_op)
            else:
                optimized.append(op)

        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if column pruning can be applied."""
        select_ops = [op for op in operations if op.type == OperationType.SELECT]
        return len(select_ops) > 0

    def _find_needed_columns(self, operations: List[Operation]) -> Set[str]:
        """Find all columns that are actually needed."""
        needed = set()

        # Start from the end and work backwards
        for op in reversed(operations):
            if op.type == OperationType.SELECT:
                needed.update(op.columns)
            elif op.type == OperationType.FILTER:
                # Add columns used in filter predicates
                for pred in op.predicates:
                    if "column" in pred:
                        needed.add(pred["column"])
            elif op.type == OperationType.GROUP_BY:
                needed.update(op.group_by_columns)
            elif op.type == OperationType.ORDER_BY:
                needed.update(op.order_by_columns)
            elif op.type == OperationType.JOIN:
                # Add columns used in join conditions
                for condition in op.join_conditions:
                    if "left_column" in condition:
                        needed.add(condition["left_column"])
                    if "right_column" in condition:
                        needed.add(condition["right_column"])

        return needed


class JoinOptimizationRule(OptimizationRule):
    """Optimize join operations for better performance."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Reorder joins and optimize join conditions."""
        if not self.can_apply(operations):
            return operations

        optimized = []
        join_ops = []
        other_ops = []

        # Separate join operations from others
        for op in operations:
            if op.type == OperationType.JOIN:
                join_ops.append(op)
            else:
                other_ops.append(op)

        # Optimize join order (simple heuristic: smaller tables first)
        if join_ops:
            optimized_joins = self._optimize_join_order(join_ops)
            optimized.extend(optimized_joins)

        optimized.extend(other_ops)
        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if join optimization can be applied."""
        join_ops = [op for op in operations if op.type == OperationType.JOIN]
        return len(join_ops) > 0

    def _optimize_join_order(self, join_ops: List[Operation]) -> List[Operation]:
        """Optimize the order of join operations."""

        # Simple heuristic: sort by estimated size (metadata)
        def get_estimated_size(op: Operation) -> int:
            return cast(int, op.metadata.get("estimated_size", 1000))  # Default size

        return sorted(join_ops, key=get_estimated_size)


class PredicatePushdownRule(OptimizationRule):
    """Push predicates down to reduce data early."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Push predicates as early as possible."""
        if not self.can_apply(operations):
            return operations

        # Collect all predicates
        all_predicates = []
        for op in operations:
            if op.predicates:
                all_predicates.extend(op.predicates)

        # Push predicates to the earliest possible operation
        optimized = []
        predicates_pushed = False

        for op in operations:
            if not predicates_pushed and self._can_push_predicates_to(op):
                # Add predicates to this operation
                optimized_op = Operation(
                    type=op.type,
                    columns=op.columns,
                    predicates=all_predicates,
                    join_conditions=op.join_conditions,
                    group_by_columns=op.group_by_columns,
                    order_by_columns=op.order_by_columns,
                    limit_count=op.limit_count,
                    window_specs=op.window_specs,
                    metadata=op.metadata,
                )
                optimized.append(optimized_op)
                predicates_pushed = True
            else:
                optimized.append(op)

        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if predicate pushdown can be applied."""
        return any(op.predicates for op in operations)

    def _can_push_predicates_to(self, op: Operation) -> bool:
        """Check if predicates can be pushed to this operation."""
        # Can push to SELECT, FILTER, but not to GROUP_BY, ORDER_BY, WINDOW
        return op.type in [OperationType.SELECT, OperationType.FILTER]


class ProjectionPushdownRule(OptimizationRule):
    """Push column projections as early as possible."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Push column selections as early as possible."""
        if not self.can_apply(operations):
            return operations

        # Find all columns that will be needed
        needed_columns = self._find_final_columns(operations)

        optimized = []
        for op in operations:
            if op.type == OperationType.SELECT:
                # Only select columns that will be needed
                optimized_op = Operation(
                    type=op.type,
                    columns=list(needed_columns),
                    predicates=op.predicates,
                    join_conditions=op.join_conditions,
                    group_by_columns=op.group_by_columns,
                    order_by_columns=op.order_by_columns,
                    limit_count=op.limit_count,
                    window_specs=op.window_specs,
                    metadata=op.metadata,
                )
                optimized.append(optimized_op)
            else:
                optimized.append(op)

        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if projection pushdown can be applied."""
        return any(op.type == OperationType.SELECT for op in operations)

    def _find_final_columns(self, operations: List[Operation]) -> Set[str]:
        """Find columns that will be in the final result."""
        # Start from the last SELECT operation
        for op in reversed(operations):
            if op.type == OperationType.SELECT:
                return set(op.columns)
        return set()


class LimitPushdownRule(OptimizationRule):
    """Push LIMIT operations as early as possible."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Push LIMIT operations to reduce data early."""
        if not self.can_apply(operations):
            return operations

        # Find the smallest LIMIT
        min_limit = None
        for op in operations:
            if op.type == OperationType.LIMIT and op.limit_count is not None:
                if min_limit is None or op.limit_count < min_limit:
                    min_limit = op.limit_count

        if min_limit is None:
            return operations

        # Push LIMIT to the earliest possible operation
        optimized = []
        limit_pushed = False

        for op in operations:
            if not limit_pushed and self._can_push_limit_to(op):
                # Add LIMIT to this operation
                optimized_op = Operation(
                    type=op.type,
                    columns=op.columns,
                    predicates=op.predicates,
                    join_conditions=op.join_conditions,
                    group_by_columns=op.group_by_columns,
                    order_by_columns=op.order_by_columns,
                    limit_count=min_limit,
                    window_specs=op.window_specs,
                    metadata=op.metadata,
                )
                optimized.append(optimized_op)
                limit_pushed = True
            else:
                optimized.append(op)

        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if limit pushdown can be applied."""
        return any(op.type == OperationType.LIMIT for op in operations)

    def _can_push_limit_to(self, op: Operation) -> bool:
        """Check if LIMIT can be pushed to this operation."""
        # Can push to SELECT, FILTER, but not to GROUP_BY, ORDER_BY, WINDOW
        return op.type in [OperationType.SELECT, OperationType.FILTER]


class UnionOptimizationRule(OptimizationRule):
    """Optimize UNION operations."""

    def apply(self, operations: List[Operation]) -> List[Operation]:
        """Optimize UNION operations by removing duplicates early."""
        if not self.can_apply(operations):
            return operations

        optimized = []
        union_ops = []
        other_ops = []

        # Separate UNION operations from others
        for op in operations:
            if op.type == OperationType.UNION:
                union_ops.append(op)
            else:
                other_ops.append(op)

        # Optimize UNION operations
        if union_ops:
            optimized_unions = self._optimize_unions(union_ops)
            optimized.extend(optimized_unions)

        optimized.extend(other_ops)
        return optimized

    def can_apply(self, operations: List[Operation]) -> bool:
        """Check if union optimization can be applied."""
        union_ops = [op for op in operations if op.type == OperationType.UNION]
        return len(union_ops) > 0

    def _optimize_unions(self, union_ops: List[Operation]) -> List[Operation]:
        """Optimize UNION operations."""
        # Simple optimization: combine consecutive UNIONs
        if len(union_ops) <= 1:
            return union_ops

        # For now, just return as-is
        # More sophisticated optimizations could be added here
        return union_ops



"""
Configuration management module for Mock Spark.

This module provides configuration management for Mock Spark,
including session configuration, runtime settings, and
environment-specific configurations.

Components:
    - Configuration: Main configuration management
    - ConfigBuilder: Configuration builder pattern
    - EnvironmentConfig: Environment-specific settings
"""

from .configuration import MockConfiguration, MockConfigBuilder, MockSparkConfig

__all__ = [
    "MockConfiguration",
    "MockConfigBuilder",
    "MockSparkConfig",
]



"""
Schema type definitions.

This module defines the abstract interfaces for schema operations,
including struct types, field definitions, and schema management.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from .data_types import IDataType


class IStructField(ABC):
    """Abstract interface for struct field operations."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Get field name."""
        pass

    @property
    @abstractmethod
    def data_type(self) -> "IDataType":
        """Get field data type."""
        pass

    @property
    @abstractmethod
    def nullable(self) -> bool:
        """Check if field is nullable."""
        pass

    @property
    @abstractmethod
    def metadata(self) -> Dict[str, Any]:
        """Get field metadata."""
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Check equality with another field."""
        pass

    @abstractmethod
    def __hash__(self) -> int:
        """Get hash for field."""
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Get string representation."""
        pass

    @abstractmethod
    def __repr__(self) -> str:
        """Get representation."""
        pass


class IStructType(ABC):
    """Abstract interface for struct type operations."""

    @property
    @abstractmethod
    def fields(self) -> List[IStructField]:
        """Get struct fields."""
        pass

    @abstractmethod
    def add_field(self, field: IStructField) -> None:
        """Add field to struct."""
        pass

    @abstractmethod
    def remove_field(self, field_name: str) -> None:
        """Remove field from struct."""
        pass

    @abstractmethod
    def get_field(self, field_name: str) -> Optional[IStructField]:
        """Get field by name."""
        pass

    @abstractmethod
    def field_names(self) -> List[str]:
        """Get field names."""
        pass

    @abstractmethod
    def field_types(self) -> Dict[str, "IDataType"]:
        """Get field types."""
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Check equality with another struct."""
        pass

    @abstractmethod
    def __hash__(self) -> int:
        """Get hash for struct."""
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Get string representation."""
        pass

    @abstractmethod
    def __repr__(self) -> str:
        """Get representation."""
        pass


class ISchema(ABC):
    """Abstract interface for schema operations."""

    @property
    @abstractmethod
    def fields(self) -> List[IStructField]:
        """Get schema fields."""
        pass

    @abstractmethod
    def add_field(self, field: IStructField) -> None:
        """Add field to schema."""
        pass

    @abstractmethod
    def remove_field(self, field_name: str) -> None:
        """Remove field from schema."""
        pass

    @abstractmethod
    def get_field(self, field_name: str) -> Optional[IStructField]:
        """Get field by name."""
        pass

    @abstractmethod
    def field_names(self) -> List[str]:
        """Get field names."""
        pass

    @abstractmethod
    def field_types(self) -> Dict[str, "IDataType"]:
        """Get field types."""
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Check equality with another schema."""
        pass

    @abstractmethod
    def __hash__(self) -> int:
        """Get hash for schema."""
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Get string representation."""
        pass

    @abstractmethod
    def __repr__(self) -> str:
        """Get representation."""
        pass


class ISchemaBuilder(ABC):
    """Abstract interface for schema building."""

    @abstractmethod
    def add_field(
        self,
        name: str,
        data_type: "IDataType",
        nullable: bool = True,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> "ISchemaBuilder":
        """Add field to schema."""
        pass

    @abstractmethod
    def add_fields(self, fields: List[IStructField]) -> "ISchemaBuilder":
        """Add multiple fields to schema."""
        pass

    @abstractmethod
    def build(self) -> IStructType:
        """Build the schema."""
        pass

    @abstractmethod
    def clear(self) -> "ISchemaBuilder":
        """Clear all fields."""
        pass


class ISchemaValidator(ABC):
    """Abstract interface for schema validation."""

    @abstractmethod
    def validate_schema(self, schema: IStructType) -> bool:
        """Validate schema structure."""
        pass

    @abstractmethod
    def validate_field(self, field: IStructField) -> bool:
        """Validate field structure."""
        pass

    @abstractmethod
    def validate_data_against_schema(
        self, data: List[Dict[str, Any]], schema: IStructType
    ) -> bool:
        """Validate data against schema."""
        pass

    @abstractmethod
    def get_validation_errors(self, schema: IStructType) -> List[str]:
        """Get validation errors."""
        pass



"""
Core session components for Mock Spark.

This module provides the core session components including the main session class,
builder pattern implementation, and context management.
"""

from .session import MockSparkSession
from .builder import MockSparkSessionBuilder
from ..context import MockSparkContext, MockJVMContext, MockJVMFunctions

__all__ = [
    "MockSparkSession",
    "MockSparkSessionBuilder",
    "MockSparkContext",
    "MockJVMContext",
    "MockJVMFunctions",
]



"""
SQL Query Builder for Mock Spark.

This module provides building of SQL queries including CTE queries,
CASE WHEN expressions, and other SQL construction utilities.
"""

from typing import Any, List, Tuple
from .cte_query_builder import CTEQueryBuilder
from .sql_expression_translator import SQLExpressionTranslator
from ...functions.core.column import MockColumnOperation


class SQLQueryBuilder:
    """Builds SQL queries for Mock Spark operations."""

    def __init__(self, cte_builder: CTEQueryBuilder):
        """Initialize the SQL query builder.

        Args:
            cte_builder: CTE query builder instance
        """
        self.cte_builder = cte_builder
        # Create a dummy table manager for the translator
        from .table_manager import DuckDBTableManager
        from sqlalchemy import create_engine, MetaData

        dummy_engine = create_engine("duckdb:///:memory:")
        dummy_metadata = MetaData()
        table_manager = DuckDBTableManager(dummy_engine, dummy_metadata)
        self.sql_converter = SQLExpressionTranslator(table_manager)

    def build_cte_query(
        self, source_table_name: str, operations: List[Tuple[str, Any]]
    ) -> str:
        """Build a single SQL query with CTEs for all operations.

        Args:
            source_table_name: Name of the source table
            operations: List of operations to apply

        Returns:
            SQL query string
        """
        return self.cte_builder.build_cte_query(source_table_name, operations)

    def build_case_when_sql(self, case_when_obj: Any, source_table_obj: Any) -> str:
        """Build SQL CASE WHEN expression from MockCaseWhen object.

        Args:
            case_when_obj: MockCaseWhen object
            source_table_obj: Source table object

        Returns:
            SQL CASE WHEN expression string
        """
        sql_parts = ["CASE"]

        # Add WHEN conditions
        for condition, value in case_when_obj.conditions:
            # Convert condition to SQL - check if it's a complex expression
            if isinstance(condition, MockColumnOperation):
                # Generate raw SQL without quoting for complex expressions
                condition_sql = self.sql_converter.expression_to_sql(condition)
            else:
                condition_sql = self.sql_converter.condition_to_sql(
                    condition, source_table_obj
                )

            # Convert value to SQL - handle MockLiteral with boolean values specially
            if hasattr(value, "value") and isinstance(value.value, bool):
                value_sql = "TRUE" if value.value else "FALSE"
            else:
                value_sql = self.sql_converter.value_to_sql(value)
            sql_parts.append(f"WHEN {condition_sql} THEN {value_sql}")

        # Add ELSE clause if default_value is set
        if case_when_obj.default_value is not None:
            # Update the ELSE clause to handle boolean MockLiterals
            if hasattr(case_when_obj.default_value, "value") and isinstance(
                case_when_obj.default_value.value, bool
            ):
                else_sql = "TRUE" if case_when_obj.default_value.value else "FALSE"
            else:
                else_sql = self.sql_converter.value_to_sql(case_when_obj.default_value)
            sql_parts.append(f"ELSE {else_sql}")

        sql_parts.append("END")
        return " ".join(sql_parts)

    def build_select_with_window_cte(
        self, source_name: str, columns: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for select with window functions.

        Args:
            source_name: Name of the source table
            columns: Columns to select
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_select_with_window_cte(
            source_name, columns, source_table_obj
        )

    def build_filter_cte(
        self, source_name: str, condition: Any, source_table_obj: Any
    ) -> str:
        """Build CTE SQL for filter operation.

        Args:
            source_name: Name of the source table
            condition: Filter condition
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_filter_cte(
            source_name, condition, source_table_obj
        )

    def build_select_cte(
        self, source_name: str, columns: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for select operation.

        Args:
            source_name: Name of the source table
            columns: Columns to select
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_select_cte(source_name, columns, source_table_obj)

    def build_with_column_cte(
        self, source_name: str, col_name: str, col: Any, source_table_obj: Any
    ) -> str:
        """Build CTE SQL for withColumn operation.

        Args:
            source_name: Name of the source table
            col_name: Name of the new column
            col: Column expression
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_with_column_cte(
            source_name, col_name, col, source_table_obj
        )

    def build_order_by_cte(
        self, source_name: str, columns: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for orderBy operation.

        Args:
            source_name: Name of the source table
            columns: Columns to order by
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_order_by_cte(
            source_name, columns, source_table_obj
        )

    def build_limit_cte(
        self, source_name: str, limit_count: int, source_table_obj: Any
    ) -> str:
        """Build CTE SQL for limit operation.

        Args:
            source_name: Name of the source table
            limit_count: Number of rows to limit
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_limit_cte(source_name, limit_count)

    def build_join_cte(
        self, source_name: str, join_params: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for join operation.

        Args:
            source_name: Name of the source table
            join_params: Join parameters
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_join_cte(
            source_name, join_params, source_table_obj
        )

    def build_union_cte(
        self, source_name: str, other_df: Any, source_table_obj: Any
    ) -> str:
        """Build CTE SQL for union operation.

        Args:
            source_name: Name of the source table
            other_df: Other DataFrame to union with
            source_table_obj: Source table object

        Returns:
            SQL query string
        """
        return self.cte_builder.build_union_cte(source_name, other_df, source_table_obj)



"""Aggregation operations for MockDataFrame."""

from typing import Any, Dict, List, Union
import statistics
from ...spark_types import MockStructType, MockStructField, StringType


class AggregationOperations:
    """Handles aggregation operations for MockDataFrame."""

    @staticmethod
    def describe(
        data: List[Dict[str, Any]], schema: MockStructType, cols: List[str]
    ) -> tuple[List[Dict[str, Any]], MockStructType]:
        """Compute basic statistics for numeric columns.

        Args:
            data: DataFrame data
            schema: DataFrame schema
            cols: Column names to describe

        Returns:
            Tuple of (result_data, result_schema)
        """
        # Determine which columns to describe
        if not cols:
            # Describe all numeric columns
            numeric_cols = []
            for field in schema.fields:
                field_type = field.dataType.typeName()
                if field_type in [
                    "long",
                    "int",
                    "integer",
                    "bigint",
                    "double",
                    "float",
                ]:
                    numeric_cols.append(field.name)
        else:
            numeric_cols = list(cols)
            # Validate that columns exist
            available_cols = [field.name for field in schema.fields]
            for col in numeric_cols:
                if col not in available_cols:
                    from ...core.exceptions.analysis import ColumnNotFoundException

                    raise ColumnNotFoundException(col)

        if not numeric_cols:
            # No numeric columns found
            return [], schema

        # Calculate statistics for each column
        result_data = []

        for col in numeric_cols:
            # Extract values for this column
            values = []
            for row in data:
                value = row.get(col)
                if value is not None and isinstance(value, (int, float)):
                    values.append(value)

            if not values:
                # No valid numeric values
                stats_row = {
                    "summary": col,
                    "count": "0",
                    "mean": "NaN",
                    "stddev": "NaN",
                    "min": "NaN",
                    "max": "NaN",
                }
            else:
                stats_row = {
                    "summary": col,
                    "count": str(len(values)),
                    "mean": str(round(statistics.mean(values), 4)),
                    "stddev": str(
                        round(statistics.stdev(values) if len(values) > 1 else 0.0, 4)
                    ),
                    "min": str(min(values)),
                    "max": str(max(values)),
                }

            result_data.append(stats_row)

        # Create result schema
        result_schema = MockStructType(
            [
                MockStructField("summary", StringType()),
                MockStructField("count", StringType()),
                MockStructField("mean", StringType()),
                MockStructField("stddev", StringType()),
                MockStructField("min", StringType()),
                MockStructField("max", StringType()),
            ]
        )

        return result_data, result_schema

    @staticmethod
    def summary(
        data: List[Dict[str, Any]], schema: MockStructType, stats: List[str]
    ) -> tuple[List[Dict[str, Any]], MockStructType]:
        """Compute extended statistics for numeric columns.

        Args:
            data: DataFrame data
            schema: DataFrame schema
            stats: Statistics to compute

        Returns:
            Tuple of (result_data, result_schema)
        """
        # Default statistics if none provided
        if not stats:
            stats = ["count", "mean", "stddev", "min", "25%", "50%", "75%", "max"]

        # Find numeric columns
        numeric_cols = []
        for field in schema.fields:
            field_type = field.dataType.typeName()
            if field_type in [
                "long",
                "int",
                "integer",
                "bigint",
                "double",
                "float",
            ]:
                numeric_cols.append(field.name)

        if not numeric_cols:
            # No numeric columns found
            return [], schema

        # Calculate statistics for each column
        result_data = []

        for col in numeric_cols:
            # Extract values for this column
            values = []
            for row in data:
                value = row.get(col)
                if value is not None and isinstance(value, (int, float)):
                    values.append(value)

            if not values:
                # No valid numeric values
                stats_row = {"summary": col}
                for stat in stats:
                    stats_row[stat] = "NaN"
            else:
                stats_row = {"summary": col}

                # Calculate requested statistics
                for stat in stats:
                    if stat == "count":
                        stats_row[stat] = str(len(values))
                    elif stat == "mean":
                        stats_row[stat] = str(round(statistics.mean(values), 4))
                    elif stat == "stddev":
                        stats_row[stat] = str(
                            round(
                                statistics.stdev(values) if len(values) > 1 else 0.0, 4
                            )
                        )
                    elif stat == "min":
                        stats_row[stat] = str(min(values))
                    elif stat == "max":
                        stats_row[stat] = str(max(values))
                    elif stat == "25%":
                        stats_row[stat] = str(
                            round(statistics.quantiles(values, n=4)[0], 4)
                        )
                    elif stat == "50%":
                        stats_row[stat] = str(round(statistics.median(values), 4))
                    elif stat == "75%":
                        stats_row[stat] = str(
                            round(statistics.quantiles(values, n=4)[2], 4)
                        )
                    else:
                        stats_row[stat] = "NaN"

            result_data.append(stats_row)

        # Create result schema
        schema_fields = [MockStructField("summary", StringType())]
        for stat in stats:
            schema_fields.append(MockStructField(stat, StringType()))

        result_schema = MockStructType(schema_fields)

        return result_data, result_schema

    @staticmethod
    def compute_basic_stats(values: List[Union[int, float]]) -> Dict[str, str]:
        """Compute basic statistics for a list of numeric values.

        Args:
            values: List of numeric values

        Returns:
            Dictionary with basic statistics
        """
        if not values:
            return {
                "count": "0",
                "mean": "NaN",
                "stddev": "NaN",
                "min": "NaN",
                "max": "NaN",
            }

        return {
            "count": str(len(values)),
            "mean": str(round(statistics.mean(values), 4)),
            "stddev": str(
                round(statistics.stdev(values) if len(values) > 1 else 0.0, 4)
            ),
            "min": str(min(values)),
            "max": str(max(values)),
        }

    @staticmethod
    def compute_extended_stats(
        values: List[Union[int, float]], stats: List[str]
    ) -> Dict[str, str]:
        """Compute extended statistics for a list of numeric values.

        Args:
            values: List of numeric values
            stats: List of statistics to compute

        Returns:
            Dictionary with requested statistics
        """
        if not values:
            result = {}
            for stat in stats:
                result[stat] = "NaN"
            return result

        result = {}
        for stat in stats:
            if stat == "count":
                result[stat] = str(len(values))
            elif stat == "mean":
                result[stat] = str(round(statistics.mean(values), 4))
            elif stat == "stddev":
                result[stat] = str(
                    round(statistics.stdev(values) if len(values) > 1 else 0.0, 4)
                )
            elif stat == "min":
                result[stat] = str(min(values))
            elif stat == "max":
                result[stat] = str(max(values))
            elif stat == "25%":
                result[stat] = str(round(statistics.quantiles(values, n=4)[0], 4))
            elif stat == "50%":
                result[stat] = str(round(statistics.median(values), 4))
            elif stat == "75%":
                result[stat] = str(round(statistics.quantiles(values, n=4)[2], 4))
            else:
                result[stat] = "NaN"

        return result

    @staticmethod
    def get_numeric_columns(schema: MockStructType) -> List[str]:
        """Get list of numeric column names from schema.

        Args:
            schema: DataFrame schema

        Returns:
            List of numeric column names
        """
        numeric_cols = []
        for field in schema.fields:
            field_type = field.dataType.typeName()
            if field_type in [
                "long",
                "int",
                "integer",
                "bigint",
                "double",
                "float",
            ]:
                numeric_cols.append(field.name)
        return numeric_cols

    @staticmethod
    def extract_numeric_values(
        data: List[Dict[str, Any]], column: str
    ) -> List[Union[int, float]]:
        """Extract numeric values from a specific column.

        Args:
            data: DataFrame data
            column: Column name

        Returns:
            List of numeric values
        """
        values = []
        for row in data:
            value = row.get(column)
            if value is not None and isinstance(value, (int, float)):
                values.append(value)
        return values



"""
Tests for PySpark version compatibility gating.

This module tests that version-specific installation correctly restricts
API access based on PySpark version compatibility mode.
"""

import pytest
from mock_spark._version_compat import (
    set_pyspark_version,
    get_pyspark_version,
    is_available,
)


class TestVersionCompatModule:
    """Test the version compatibility module itself."""

    def test_set_and_get_version(self):
        """Test setting and getting PySpark version."""
        original = get_pyspark_version()
        try:
            set_pyspark_version("3.1")
            assert get_pyspark_version() == "3.1.3"

            set_pyspark_version("3.5")
            assert get_pyspark_version() == "3.5.2"
        finally:
            # Reset
            set_pyspark_version(original) if original else set_pyspark_version(None)  # type: ignore[arg-type]

    def test_version_normalization(self):
        """Test that version strings are normalized correctly."""
        original = get_pyspark_version()
        try:
            set_pyspark_version("3.2")
            assert get_pyspark_version() == "3.2.4"

            set_pyspark_version("3.2.4")
            assert get_pyspark_version() == "3.2.4"
        finally:
            set_pyspark_version(original) if original else set_pyspark_version(None)  # type: ignore[arg-type]


class TestFunctionVersionGating:
    """Test that functions are gated by version correctly."""

    def setup_method(self):
        """Save original version before each test."""
        self.original_version = get_pyspark_version()

    def teardown_method(self):
        """Restore original version after each test."""
        if self.original_version:
            set_pyspark_version(self.original_version)
        else:
            # Reset to None for all features
            from mock_spark import _version_compat

            _version_compat._PYSPARK_COMPAT_VERSION = None

    @pytest.mark.skip(
        reason="Direct imports bypass version gating - only F.function_name access is gated"
    )
    def test_make_date_unavailable_in_30(self):
        """Test that make_date raises AttributeError in 3.0 mode."""
        set_pyspark_version("3.0")

        # make_date added in PySpark 3.3
        with pytest.raises(AttributeError, match="has no attribute 'make_date'"):
            from mock_spark.functions import make_date  # noqa: F401

    def test_make_date_available_in_33(self):
        """Test that make_date is available in 3.3 mode."""
        set_pyspark_version("3.3")

        # Should not raise
        from mock_spark.functions import make_date

        assert make_date is not None

    @pytest.mark.skip(
        reason="Direct imports bypass version gating - only F.function_name access is gated"
    )
    def test_acosh_unavailable_in_30(self):
        """Test that acosh raises AttributeError in 3.0 mode."""
        set_pyspark_version("3.0")

        # acosh added in PySpark 3.1
        with pytest.raises(AttributeError, match="has no attribute 'acosh'"):
            from mock_spark.functions import acosh  # noqa: F401

    def test_acosh_available_in_31(self):
        """Test that acosh is available in 3.1 mode."""
        set_pyspark_version("3.1")

        # Should not raise
        from mock_spark.functions import acosh

        assert acosh is not None

    @pytest.mark.skip(
        reason="Direct imports bypass version gating - only F.function_name access is gated"
    )
    def test_bool_and_unavailable_in_34(self):
        """Test that bool_and raises AttributeError in 3.4 mode."""
        set_pyspark_version("3.4")

        # bool_and added in PySpark 3.5
        with pytest.raises(AttributeError, match="has no attribute 'bool_and'"):
            from mock_spark.functions import bool_and  # noqa: F401

    def test_bool_and_available_in_35(self):
        """Test that bool_and is available in 3.5 mode."""
        set_pyspark_version("3.5")

        # Should not raise
        from mock_spark.functions import bool_and

        assert bool_and is not None

    def test_all_functions_available_default(self):
        """Test that all functions available when no version set."""
        # Reset to default (no version)
        from mock_spark import _version_compat

        _version_compat._PYSPARK_COMPAT_VERSION = None

        # All should be available
        from mock_spark.functions import make_date, acosh, bool_and

        assert make_date is not None
        assert acosh is not None
        assert bool_and is not None

    def test_overlay_available_in_30(self):
        """Test that overlay is available in 3.0 (it exists in PySpark 3.0)."""
        set_pyspark_version("3.0")

        # overlay exists in PySpark 3.0
        from mock_spark.functions import overlay

        assert overlay is not None


class TestDataFrameMethodGating:
    """Test that DataFrame methods are gated by version correctly."""

    def setup_method(self):
        """Save original version before each test."""
        self.original_version = get_pyspark_version()

    def teardown_method(self):
        """Restore original version after each test."""
        if self.original_version:
            set_pyspark_version(self.original_version)
        else:
            # Reset to None for all features
            from mock_spark import _version_compat

            _version_compat._PYSPARK_COMPAT_VERSION = None

    def test_dataframe_basic_methods_always_available(self):
        """Test that basic DataFrame methods are always available."""
        from mock_spark import MockSparkSession

        set_pyspark_version("3.0")
        spark = MockSparkSession("test")
        df = spark.createDataFrame([{"id": 1}])

        # Core methods should always be available
        assert hasattr(df, "select")
        assert hasattr(df, "filter")
        assert hasattr(df, "groupBy")
        assert hasattr(df, "show")

    def test_unpivot_unavailable_in_30(self):
        """Test that unpivot raises AttributeError in 3.0 mode."""
        from mock_spark import MockSparkSession

        set_pyspark_version("3.0")
        spark = MockSparkSession("test")
        df = spark.createDataFrame([{"id": 1, "a": 2, "b": 3}])

        # unpivot added in PySpark 3.4
        with pytest.raises(AttributeError, match="has no attribute 'unpivot'"):
            df.unpivot(ids="id", values=["a", "b"])

    def test_unpivot_available_in_34(self):
        """Test that unpivot is available in 3.4 mode."""
        from mock_spark import MockSparkSession

        set_pyspark_version("3.4")
        spark = MockSparkSession("test")
        df = spark.createDataFrame([{"id": 1, "a": 2, "b": 3}])

        # Should not raise (though may fail for other reasons)
        assert hasattr(df, "unpivot")

    def test_transform_availability(self):
        """Test transform method availability across versions."""
        from mock_spark import MockSparkSession

        # transform added in PySpark 3.0
        set_pyspark_version("3.0")
        spark = MockSparkSession("test")
        df = spark.createDataFrame([{"id": 1}])

        assert hasattr(df, "transform")


class TestIsAvailableFunction:
    """Test the is_available() function directly."""

    def setup_method(self):
        """Save original version before each test."""
        self.original_version = get_pyspark_version()

    def teardown_method(self):
        """Restore original version after each test."""
        if self.original_version:
            set_pyspark_version(self.original_version)
        else:
            from mock_spark import _version_compat

            _version_compat._PYSPARK_COMPAT_VERSION = None

    def test_is_available_default_mode(self):
        """Test that is_available returns True for all items in default mode."""
        from mock_spark import _version_compat

        _version_compat._PYSPARK_COMPAT_VERSION = None

        assert is_available("make_date", "function") is True
        assert is_available("acosh", "function") is True
        assert is_available("nonexistent", "function") is True  # Unknown items allowed

    def test_is_available_version_30(self):
        """Test is_available in PySpark 3.0 mode."""
        set_pyspark_version("3.0")

        # Functions added after 3.0 should not be available
        assert is_available("acosh", "function") is False
        assert is_available("make_date", "function") is False
        assert is_available("bool_and", "function") is False

        # Functions in 3.0 should be available
        assert is_available("col", "function") is True
        assert is_available("upper", "function") is True
        assert is_available("overlay", "function") is True

    def test_is_available_version_35(self):
        """Test is_available in PySpark 3.5 mode."""
        set_pyspark_version("3.5")

        # All tested functions should be available in 3.5
        assert is_available("acosh", "function") is True
        assert is_available("make_date", "function") is True
        assert is_available("bool_and", "function") is True
        assert is_available("col", "function") is True

    def test_is_available_dataframe_methods(self):
        """Test is_available for DataFrame methods."""
        set_pyspark_version("3.0")

        # Core methods should be available
        assert is_available("select", "dataframe_method") is True
        assert is_available("filter", "dataframe_method") is True

        # Methods added later should not be available
        # (unpivot added in 3.4)
        assert is_available("unpivot", "dataframe_method") is False


class TestVersionGatingIntegration:
    """Integration tests for version gating."""

    def setup_method(self):
        """Save original version before each test."""
        self.original_version = get_pyspark_version()

    def teardown_method(self):
        """Restore original version after each test."""
        if self.original_version:
            set_pyspark_version(self.original_version)
        else:
            from mock_spark import _version_compat

            _version_compat._PYSPARK_COMPAT_VERSION = None

    @pytest.mark.skip(
        reason="F.function_name access gating not fully implemented - deferred"
    )
    def test_version_31_full_workflow(self):
        """Test complete workflow in PySpark 3.1 compatibility mode."""
        from mock_spark import MockSparkSession
        import mock_spark.functions as F

        set_pyspark_version("3.1")
        spark = MockSparkSession("test")

        # Create DataFrame
        data = [{"id": 1, "value": 2.5}, {"id": 2, "value": 3.7}]
        df = spark.createDataFrame(data)

        # Use functions available in 3.1
        result = df.select(
            F.col("id"),
            F.acosh(F.lit(2)).alias("acosh_result"),  # acosh added in 3.1
        )

        assert result is not None

        # Try to use function not in 3.1 (make_date added in 3.3)
        with pytest.raises(AttributeError):
            df.select(F.make_date(F.lit(2024), F.lit(1), F.lit(1)))

    @pytest.mark.skip(
        reason="F.function_name access gating not fully implemented - deferred"
    )
    def test_version_33_full_workflow(self):
        """Test complete workflow in PySpark 3.3 compatibility mode."""
        from mock_spark import MockSparkSession
        import mock_spark.functions as F

        set_pyspark_version("3.3")
        spark = MockSparkSession("test")

        # Create DataFrame
        data = [{"year": 2024, "month": 3, "day": 15}]
        df = spark.createDataFrame(data)

        # Use functions available in 3.3
        result = df.select(
            F.make_date(F.col("year"), F.col("month"), F.col("day")).alias("date")
        )

        assert result is not None

        # bool_and added in 3.5, should fail
        with pytest.raises(AttributeError):
            df.select(F.bool_and(F.lit(True)))

    def test_switching_versions(self):
        """Test that version can be switched multiple times."""
        assert is_available("make_date", "function")  # Default: available

        set_pyspark_version("3.0")
        assert not is_available("make_date", "function")  # 3.0: not available

        set_pyspark_version("3.3")
        assert is_available("make_date", "function")  # 3.3: available

        set_pyspark_version("3.0")
        assert not is_available("make_date", "function")  # Back to 3.0: not available



"""
CTE (Common Table Expression) query builder for Mock Spark.

This module provides CTE query construction for complex DataFrame operations.
"""

from typing import Any, List, Tuple
from .sql_expression_translator import SQLExpressionTranslator


class CTEQueryBuilder:
    """Builds CTE (Common Table Expression) queries for DataFrame operations."""

    def __init__(self, expression_translator: SQLExpressionTranslator):
        """Initialize CTE query builder.

        Args:
            expression_translator: SQL expression translator for converting expressions
        """
        self.expression_translator = expression_translator

    def build_cte_query(
        self, source_table_name: str, operations: List[Tuple[str, Any]]
    ) -> str:
        """Build a single SQL query with CTEs for all operations.

        Args:
            source_table_name: Name of the initial table with data
            operations: List of (operation_name, operation_payload) tuples

        Returns:
            Complete SQL query with CTEs
        """
        # Get source table from table manager
        table_manager = self.expression_translator.table_manager
        source_table_obj = table_manager.get_table(source_table_name)
        if source_table_obj is None:
            raise ValueError(f"Table {source_table_name} not found")

        cte_definitions = []
        current_cte_name = source_table_name
        # Track columns as we build CTEs for operations that modify schema
        current_columns = [c.name for c in source_table_obj.columns]

        for i, (op_name, op_val) in enumerate(operations):
            cte_name = f"cte_{i}"

            if op_name == "filter":
                cte_sql = self.build_filter_cte(
                    current_cte_name, op_val, source_table_obj
                )
            elif op_name == "select":
                cte_sql = self.build_select_cte(
                    current_cte_name, op_val, source_table_obj
                )
                # Update current columns for select operations
                # Note: This is a simplification - full implementation would parse the select
                # For now, we'll just use source columns as we don't modify in place
            elif op_name == "withColumn":
                col_name, col = op_val
                cte_sql = self.build_with_column_cte(
                    current_cte_name, col_name, col, current_columns
                )
                # Track the new column
                if col_name not in current_columns:
                    current_columns.append(col_name)
            elif op_name == "orderBy":
                cte_sql = self.build_order_by_cte(
                    current_cte_name, op_val, source_table_obj
                )
            elif op_name == "limit":
                cte_sql = self.build_limit_cte(current_cte_name, op_val)
            elif op_name == "join":
                cte_sql = self.build_join_cte(
                    current_cte_name, op_val, source_table_obj
                )
            elif op_name == "union":
                cte_sql = self.build_union_cte(
                    current_cte_name, op_val, source_table_obj
                )
            else:
                # Unknown operation, skip
                continue

            cte_definitions.append(f"{cte_name} AS ({cte_sql})")
            current_cte_name = cte_name

        # Build final query
        if cte_definitions:
            cte_clause = "WITH " + ",\n     ".join(cte_definitions)
            final_query = f"{cte_clause}\nSELECT * FROM {current_cte_name}"
        else:
            final_query = f"SELECT * FROM {source_table_name}"

        return final_query

    def build_filter_cte(
        self, source_name: str, condition: Any, source_table_obj: Any
    ) -> str:
        """Build CTE SQL for filter operation.

        Args:
            source_name: Name of the source CTE/table
            condition: Filter condition
            source_table_obj: Source table object

        Returns:
            SQL string for filter CTE
        """
        # Convert condition to SQL
        filter_sql = self.expression_translator.condition_to_sql(
            condition, source_table_obj
        )
        return f"SELECT * FROM {source_name} WHERE {filter_sql}"

    def build_select_cte(
        self, source_name: str, columns: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for select operation.

        Args:
            source_name: Name of the source CTE/table
            columns: Columns to select
            source_table_obj: Source table object

        Returns:
            SQL string for select CTE
        """
        # Check for window functions
        has_window_functions = any(
            (hasattr(col, "function_name") and hasattr(col, "window_spec"))
            or (
                hasattr(col, "function_name")
                and hasattr(col, "column")
                and col.__class__.__name__ == "MockAggregateFunction"
            )
            for col in columns
        )

        if has_window_functions:
            return self.build_select_with_window_cte(
                source_name, columns, source_table_obj
            )

        # Check for explode operations - need special handling
        explode_cols = []
        non_explode_parts = []

        for col in columns:
            is_explode = hasattr(col, "operation") and col.operation in [
                "explode",
                "explode_outer",
            ]

            if is_explode:
                explode_cols.append(col)
            elif isinstance(col, str):
                if col == "*":
                    non_explode_parts.append("*")
                else:
                    non_explode_parts.append(f'"{col}"')
            elif hasattr(col, "value") and hasattr(col, "data_type"):
                # MockLiteral
                if isinstance(col.value, str):
                    non_explode_parts.append(f"'{col.value}' AS \"{col.name}\"")
                else:
                    non_explode_parts.append(f'{col.value} AS "{col.name}"')
            elif hasattr(col, "operation"):
                # Column operation (but not explode)
                expr_sql = self.expression_translator.expression_to_sql(col)
                col_name = getattr(col, "name", "result")
                non_explode_parts.append(f'{expr_sql} AS "{col_name}"')
            elif hasattr(col, "name"):
                # Check for alias or regular column
                original_col = getattr(col, "_original_column", None) or getattr(
                    col, "original_column", None
                )
                if original_col is not None:
                    non_explode_parts.append(f'"{original_col.name}" AS "{col.name}"')
                elif col.name == "*":
                    non_explode_parts.append("*")
                else:
                    # Regular column reference (e.g., MockColumn)
                    non_explode_parts.append(f'"{col.name}"')

        # Handle explode operations - use simpler UNNEST pattern
        if explode_cols:
            # For explode, we need to select other columns plus the UNNEST result
            columns_list = ", ".join(non_explode_parts) if non_explode_parts else "*"

            # For each explode column, create UNNEST in SELECT
            explode_parts = []
            for explode_col in explode_cols:
                array_col = explode_col.column.name
                alias = getattr(explode_col, "name", "exploded")

                if explode_col.operation == "explode_outer":
                    unnest_expr = f'UNNEST(COALESCE("{array_col}", [NULL]))'
                else:
                    unnest_expr = f'UNNEST("{array_col}")'

                # Use UNNEST as scalar subquery: SELECT ..., UNNEST(...) AS col FROM table
                # Note: This syntax doesn't expand rows, it just returns the array as-is
                # We need to detect this and handle it differently - for now, return
                # a placeholder that will be handled in _apply_select for table-based execution
                explode_parts.append(f'{unnest_expr} AS "{alias}"')

            all_parts = [columns_list] if columns_list else []
            all_parts.extend(explode_parts)
            return f"SELECT {', '.join(all_parts)} FROM {source_name}"
        else:
            # No explode operations - build regular SELECT
            if non_explode_parts.count("*") > 1:
                non_explode_parts = ["*"]

            columns_clause = ", ".join(non_explode_parts) if non_explode_parts else "*"
            return f"SELECT {columns_clause} FROM {source_name}"

    def build_select_with_window_cte(
        self, source_name: str, columns: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for select with window functions.

        Args:
            source_name: Name of the source CTE/table
            columns: Columns to select
            source_table_obj: Source table object

        Returns:
            SQL string for select with window functions CTE
        """
        select_parts = []

        # Track aliases before adding window functions
        existing_cols = {}
        for col in columns:
            if (
                hasattr(col, "name")
                and hasattr(col, "_original_column")
                and col._original_column is not None
            ):
                existing_cols[col.name] = col._original_column.name

        # Handle wildcard columns first - if we have "*", include all source table columns
        has_wildcard = any(
            (isinstance(col, str) and col == "*")
            or (hasattr(col, "name") and col.name == "*")
            for col in columns
        )

        if (
            has_wildcard
            and source_table_obj is not None
            and hasattr(source_table_obj, "columns")
        ):
            # Add all source table columns when we have "*"
            for col in source_table_obj.columns:
                select_parts.append(f'"{col.name}"')

        for col in columns:
            if isinstance(col, str):
                if col == "*":
                    # Already handled above
                    continue
                else:
                    # Only add if not already included
                    if f'"{col}"' not in select_parts:
                        select_parts.append(f'"{col}"')
            elif hasattr(col, "name") and col.name == "*":
                # Already handled above
                continue
            elif hasattr(col, "function_name") and hasattr(col, "window_spec"):
                # Window function
                func_name = col.function_name.upper()
                if (
                    col.column is None
                    or col.column == "*"
                    or (hasattr(col.column, "name") and col.column.name == "*")
                ):
                    # For window functions, we don't need a column expression for functions like ROW_NUMBER
                    if func_name in ["ROW_NUMBER", "RANK", "DENSE_RANK"]:
                        col_expr = ""
                    elif func_name in ["LAG", "LEAD"]:
                        # LAG and LEAD need a specific column, use the first available column
                        if source_table_obj is not None and hasattr(
                            source_table_obj, "columns"
                        ):
                            try:
                                columns_list = list(source_table_obj.columns)
                                if columns_list:
                                    col_expr = f'"{columns_list[0].name}"'
                                else:
                                    col_expr = "*"
                            except Exception:
                                col_expr = "*"
                        else:
                            col_expr = "*"
                    else:
                        col_expr = "*"
                else:
                    if hasattr(col.column, "name"):
                        col_expr = f'"{col.column.name}"'
                    else:
                        col_expr = f'"{col.column}"'

                window_sql = self.expression_translator.window_spec_to_sql(
                    col.window_spec, source_table_obj, existing_cols
                )
                result_name = getattr(col, "name", f"{func_name.lower()}_result")
                if col_expr:
                    select_parts.append(
                        f'{func_name}({col_expr}) OVER ({window_sql}) AS "{result_name}"'
                    )
                else:
                    select_parts.append(
                        f'{func_name}() OVER ({window_sql}) AS "{result_name}"'
                    )
            elif (
                hasattr(col, "function_name")
                and col.__class__.__name__ == "MockAggregateFunction"
            ):
                # Aggregate function
                func_name = col.function_name.upper()
                if col.column is None or col.column == "*":
                    col_expr = "*"
                elif isinstance(col.column, str):
                    col_expr = f'"{col.column}"'
                else:
                    col_expr = f'"{col.column.name}"'
                result_name = getattr(col, "name", f"{func_name.lower()}_result")
                select_parts.append(f'{func_name}({col_expr}) AS "{result_name}"')
            elif hasattr(col, "name"):
                # Check if this is an aliased column
                if (
                    hasattr(col, "_original_column")
                    and col._original_column is not None
                ):
                    # This is an aliased column - use original column name with alias
                    original_name = col._original_column.name
                    alias_name = col.name
                    select_parts.append(f'"{original_name}" AS "{alias_name}"')
                else:
                    # Regular column
                    select_parts.append(f'"{col.name}"')

        columns_clause = ", ".join(select_parts) if select_parts else "*"
        return f"SELECT {columns_clause} FROM {source_name}"

    def build_with_column_cte(
        self, source_name: str, col_name: str, col: Any, existing_columns: List[str]
    ) -> str:
        """Build CTE SQL for withColumn operation.

        Args:
            source_name: Name of the source CTE/table
            col_name: Name of the column to add/replace
            col: Column expression
            existing_columns: List of column names in the source

        Returns:
            SQL string for withColumn CTE
        """
        # Check if we're replacing an existing column or adding a new one
        select_parts = []
        column_added = False

        for existing_col in existing_columns:
            if existing_col == col_name:
                # Replace this column with new expression
                if hasattr(col, "value") and hasattr(col, "data_type"):
                    # Literal value
                    if isinstance(col.value, str):
                        select_parts.append(f"'{col.value}' AS \"{col_name}\"")
                    else:
                        select_parts.append(f'{col.value} AS "{col_name}"')
                elif hasattr(col, "operation"):
                    # Column operation
                    expr_sql = self.expression_translator.expression_to_sql(col)
                    select_parts.append(f'{expr_sql} AS "{col_name}"')
                else:
                    # Simple column reference
                    select_parts.append(f'"{col_name}"')
                column_added = True
            else:
                # Keep existing column
                select_parts.append(f'"{existing_col}"')

        # If column wasn't replaced, add it as new
        if not column_added:
            if hasattr(col, "value") and hasattr(col, "data_type"):
                # Literal value
                if isinstance(col.value, str):
                    select_parts.append(f"'{col.value}' AS \"{col_name}\"")
                else:
                    select_parts.append(f'{col.value} AS "{col_name}"')
            elif hasattr(col, "operation"):
                # Column operation
                expr_sql = self.expression_translator.expression_to_sql(col)
                select_parts.append(f'{expr_sql} AS "{col_name}"')
            else:
                # Simple column reference
                select_parts.append(f'"{col_name}"')

        columns_clause = ", ".join(select_parts)
        return f"SELECT {columns_clause} FROM {source_name}"

    def build_order_by_cte(
        self, source_name: str, columns: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for orderBy operation.

        Args:
            source_name: Name of the source CTE/table
            columns: Order by columns
            source_table_obj: Source table object

        Returns:
            SQL string for orderBy CTE
        """
        order_parts = []

        for col in columns:
            if isinstance(col, str):
                order_parts.append(f'"{col}"')
            elif hasattr(col, "operation") and col.operation == "desc":
                order_parts.append(f'"{col.column.name}" DESC')
            elif hasattr(col, "operation") and col.operation == "asc":
                order_parts.append(f'"{col.column.name}" ASC')
            elif hasattr(col, "name"):
                order_parts.append(f'"{col.name}"')

        order_clause = ", ".join(order_parts)
        return f"SELECT * FROM {source_name} ORDER BY {order_clause}"

    def build_limit_cte(self, source_name: str, limit_count: int) -> str:
        """Build CTE SQL for limit operation.

        Args:
            source_name: Name of the source CTE/table
            limit_count: Number of rows to limit

        Returns:
            SQL string for limit CTE
        """
        return f"SELECT * FROM {source_name} LIMIT {limit_count}"

    def build_join_cte(
        self, source_name: str, join_params: Tuple[Any, ...], source_table_obj: Any
    ) -> str:
        """Build CTE SQL for join operation.

        Args:
            source_name: Name of the source CTE/table
            join_params: Join parameters (other_df, on, how)
            source_table_obj: Source table object

        Returns:
            SQL string for join CTE
        """
        other_df, on, how = join_params

        # Materialize the other DataFrame if it's lazy
        if hasattr(other_df, "_operations_queue") and other_df._operations_queue:
            other_df = other_df._materialize_if_lazy()

        # Create a temporary table for the other DataFrame
        table_manager = self.expression_translator.table_manager
        other_table_name = f"temp_join_{id(other_df)}"
        table_manager.create_table_with_data(other_table_name, other_df.data)

        # Build join condition
        if isinstance(on, str):
            join_condition = f'{source_name}."{on}" = {other_table_name}."{on}"'
        elif isinstance(on, list):
            conditions = [
                f'{source_name}."{col}" = {other_table_name}."{col}"' for col in on
            ]
            join_condition = " AND ".join(conditions)
        elif hasattr(on, "operation") and on.operation == "==":
            # Handle MockColumnOperation (e.g., emp_df.dept_id == dept_df.dept_id)
            # Extract column names from equality condition
            left_col = on.column.name if hasattr(on.column, "name") else str(on.column)
            # For join conditions with two different DataFrames, use the column name
            # from both sides (they should be the same in most cases)
            join_condition = (
                f'{source_name}."{left_col}" = {other_table_name}."{left_col}"'
            )
        elif hasattr(on, "operation"):
            # Other column operation as join condition - fallback
            join_condition = self.expression_translator.condition_to_sql(
                on, source_table_obj
            )
        else:
            join_condition = "1=1"  # Fallback

        # Map join type
        join_type_map = {
            "inner": "INNER JOIN",
            "left": "LEFT JOIN",
            "right": "RIGHT JOIN",
            "outer": "FULL OUTER JOIN",
            "full": "FULL OUTER JOIN",
            "left_outer": "LEFT JOIN",
            "right_outer": "RIGHT JOIN",
            "full_outer": "FULL OUTER JOIN",
        }
        join_type = join_type_map.get(how, "INNER JOIN")

        # Explicitly select columns to avoid ambiguous column names
        # Get columns from source table
        source_columns = [
            f'{source_name}."{col.name}"' for col in source_table_obj.columns
        ]
        # Get columns from other table (excluding join keys to avoid duplication)
        other_schema = other_df.schema if hasattr(other_df, "schema") else other_df
        other_column_names = [field.name for field in other_schema.fields]

        # Determine join keys for exclusion
        if isinstance(on, str):
            join_keys = [on]
        elif isinstance(on, list):
            join_keys = on
        else:
            join_keys = []

        other_columns = [
            f'{other_table_name}."{col}"'
            for col in other_column_names
            if col not in join_keys
        ]

        # Combine columns: source columns first, then other columns
        all_columns = source_columns + other_columns
        select_clause = ", ".join(all_columns)

        return f"SELECT {select_clause} FROM {source_name} {join_type} {other_table_name} ON {join_condition}"

    def build_union_cte(
        self, source_name: str, other_df: Any, source_table_obj: Any
    ) -> str:
        """Build CTE SQL for union operation.

        Args:
            source_name: Name of the source CTE/table
            other_df: Other DataFrame to union with
            source_table_obj: Source table object

        Returns:
            SQL string for union CTE
        """
        # Materialize the other DataFrame if it's lazy
        if hasattr(other_df, "_operations_queue") and other_df._operations_queue:
            other_df = other_df._materialize_if_lazy()

        # Create a temporary table for the other DataFrame
        table_manager = self.expression_translator.table_manager
        other_table_name = f"temp_union_{id(other_df)}"
        table_manager.create_table_with_data(other_table_name, other_df.data)

        return f"SELECT * FROM {source_name} UNION ALL SELECT * FROM {other_table_name}"



"""
Window function handler for MockDataFrame.

This module handles window function evaluation (row_number, rank, lag, lead, etc.)
following the Single Responsibility Principle.
"""

from typing import List, Dict, Any, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    pass


class WindowFunctionHandler:
    """Handles window function evaluation (row_number, rank, lag, lead, etc.)."""

    def __init__(self, dataframe: Any):
        """Initialize window function handler.

        Args:
            dataframe: The MockDataFrame instance this handler belongs to
        """
        self.dataframe = dataframe

    def evaluate_window_functions(
        self, data: List[Dict[str, Any]], window_functions: List[Tuple[Any, ...]]
    ) -> List[Dict[str, Any]]:
        """Evaluate window functions across all rows."""
        result_data = data.copy()

        for col_index, window_func in window_functions:
            col_name = window_func.name

            if window_func.function_name == "row_number":
                # For row_number(), we need to handle partitionBy and orderBy
                if hasattr(window_func, "window_spec") and window_func.window_spec:
                    window_spec = window_func.window_spec

                    # Get partition by columns from window spec
                    partition_by_cols = getattr(window_spec, "_partition_by", [])
                    # Get order by columns from window spec
                    order_by_cols = getattr(window_spec, "_order_by", [])

                    if partition_by_cols:
                        # Handle partitioning - group by partition columns
                        partition_groups: Dict[Any, List[int]] = {}
                        for i, row in enumerate(result_data):
                            # Create partition key
                            partition_key = tuple(
                                (
                                    row.get(col.name)
                                    if hasattr(col, "name")
                                    else row.get(str(col))
                                )
                                for col in partition_by_cols
                            )
                            if partition_key not in partition_groups:
                                partition_groups[partition_key] = []
                            partition_groups[partition_key].append(i)

                        # Assign row numbers within each partition
                        for partition_indices in partition_groups.values():
                            if order_by_cols:
                                # Sort within partition by order by columns using corrected ordering logic
                                sorted_partition_indices = (
                                    self._apply_ordering_to_indices(
                                        result_data,
                                        partition_indices,
                                        order_by_cols,
                                    )
                                )
                            else:
                                # No order by - use original order within partition
                                sorted_partition_indices = partition_indices

                            # Assign row numbers starting from 1 within each partition
                            for i, original_index in enumerate(
                                sorted_partition_indices
                            ):
                                result_data[original_index][col_name] = i + 1
                    elif order_by_cols:
                        # No partitioning, just sort by order by columns using corrected ordering logic
                        sorted_indices = self._apply_ordering_to_indices(
                            result_data, list(range(len(result_data))), order_by_cols
                        )

                        # Assign row numbers based on sorted order
                        for i, original_index in enumerate(sorted_indices):
                            result_data[original_index][col_name] = i + 1
                    else:
                        # No partition or order by - just assign sequential row numbers
                        for i in range(len(result_data)):
                            result_data[i][col_name] = i + 1
                else:
                    # No window spec - assign sequential row numbers
                    for i in range(len(result_data)):
                        result_data[i][col_name] = i + 1
            elif window_func.function_name == "lag":
                # Handle lag function - get previous row value
                self._evaluate_lag_lead(
                    result_data, window_func, col_name, is_lead=False
                )
            elif window_func.function_name == "lead":
                # Handle lead function - get next row value
                self._evaluate_lag_lead(
                    result_data, window_func, col_name, is_lead=True
                )
            elif window_func.function_name in ["rank", "dense_rank"]:
                # Handle rank and dense_rank functions
                self._evaluate_rank_functions(result_data, window_func, col_name)
            elif window_func.function_name in [
                "avg",
                "sum",
                "count",
                "countDistinct",
                "max",
                "min",
            ]:
                # Handle aggregate window functions
                self._evaluate_aggregate_window_functions(
                    result_data, window_func, col_name
                )
            else:
                # For other window functions, assign None for now
                for row in result_data:
                    row[col_name] = None

        return result_data

    def _evaluate_lag_lead(
        self, data: List[Dict[str, Any]], window_func: Any, col_name: str, is_lead: bool
    ) -> None:
        """Evaluate lag or lead window function."""
        if not window_func.column_name:
            # No column specified, set to None
            for row in data:
                row[col_name] = None
            return

        # Get offset and default value
        offset = getattr(window_func, "offset", 1)
        default_value = getattr(window_func, "default_value", None)

        # Handle window specification if present
        if hasattr(window_func, "window_spec") and window_func.window_spec:
            window_spec = window_func.window_spec
            partition_by_cols = getattr(window_spec, "_partition_by", [])
            order_by_cols = getattr(window_spec, "_order_by", [])

            if partition_by_cols:
                # Handle partitioning
                partition_groups: Dict[Any, List[int]] = {}
                for i, row in enumerate(data):
                    partition_key = tuple(
                        row.get(col.name) if hasattr(col, "name") else row.get(str(col))
                        for col in partition_by_cols
                    )
                    if partition_key not in partition_groups:
                        partition_groups[partition_key] = []
                    partition_groups[partition_key].append(i)

                # Process each partition
                for partition_indices in partition_groups.values():
                    # Apply ordering to partition indices
                    ordered_indices = self._apply_ordering_to_indices(
                        data, partition_indices, order_by_cols
                    )
                    self._apply_lag_lead_to_partition(
                        data,
                        ordered_indices,
                        window_func.column_name,
                        col_name,
                        offset,
                        default_value,
                        is_lead,
                    )
            else:
                # No partitioning, apply to entire dataset with ordering
                all_indices = list(range(len(data)))
                ordered_indices = self._apply_ordering_to_indices(
                    data, all_indices, order_by_cols
                )
                self._apply_lag_lead_to_partition(
                    data,
                    ordered_indices,
                    window_func.column_name,
                    col_name,
                    offset,
                    default_value,
                    is_lead,
                )
        else:
            # No window spec, apply to entire dataset
            self._apply_lag_lead_to_partition(
                data,
                list(range(len(data))),
                window_func.column_name,
                col_name,
                offset,
                default_value,
                is_lead,
            )

    def _apply_ordering_to_indices(
        self, data: List[Dict[str, Any]], indices: List[int], order_by_cols: List[Any]
    ) -> List[int]:
        """Apply ordering to a list of indices based on order by columns."""
        if not order_by_cols:
            return indices

        def sort_key(idx: int) -> Tuple[Any, ...]:
            row = data[idx]
            key_values = []
            for col in order_by_cols:
                # Handle MockColumnOperation objects (like col("salary").desc())
                if hasattr(col, "column") and hasattr(col.column, "name"):
                    col_name = col.column.name
                elif hasattr(col, "name"):
                    col_name = col.name
                else:
                    col_name = str(col)
                value = row.get(col_name)
                # Handle None values for sorting - put them at the end
                if value is None:
                    key_values.append(float("inf"))  # Sort None values last
                else:
                    key_values.append(value)
            return tuple(key_values)

        # Check if any column has desc operation
        has_desc = any(
            hasattr(col, "operation") and col.operation == "desc"
            for col in order_by_cols
        )

        # Sort indices based on the ordering
        return sorted(indices, key=sort_key, reverse=has_desc)

    def _apply_lag_lead_to_partition(
        self,
        data: List[Dict[str, Any]],
        indices: List[int],
        source_col: str,
        target_col: str,
        offset: int,
        default_value: Any,
        is_lead: bool,
    ) -> None:
        """Apply lag or lead to a specific partition."""
        if is_lead:
            # Lead: get next row value
            for i, idx in enumerate(indices):
                source_idx = i + offset
                if source_idx < len(indices):
                    actual_idx = indices[source_idx]
                    data[idx][target_col] = data[actual_idx].get(source_col)
                else:
                    data[idx][target_col] = default_value
        else:
            # Lag: get previous row value
            for i, idx in enumerate(indices):
                source_idx = i - offset
                if source_idx >= 0:
                    actual_idx = indices[source_idx]
                    data[idx][target_col] = data[actual_idx].get(source_col)
                else:
                    data[idx][target_col] = default_value

    def _evaluate_rank_functions(
        self, data: List[Dict[str, Any]], window_func: Any, col_name: str
    ) -> None:
        """Evaluate rank or dense_rank window function."""
        is_dense = window_func.function_name == "dense_rank"

        # Handle window specification if present
        if hasattr(window_func, "window_spec") and window_func.window_spec:
            window_spec = window_func.window_spec
            partition_by_cols = getattr(window_spec, "_partition_by", [])
            order_by_cols = getattr(window_spec, "_order_by", [])

            if partition_by_cols:
                # Handle partitioning
                partition_groups: Dict[Any, List[int]] = {}
                for i, row in enumerate(data):
                    partition_key = tuple(
                        row.get(col.name) if hasattr(col, "name") else row.get(str(col))
                        for col in partition_by_cols
                    )
                    if partition_key not in partition_groups:
                        partition_groups[partition_key] = []
                    partition_groups[partition_key].append(i)

                # Process each partition
                for partition_indices in partition_groups.values():
                    self._apply_rank_to_partition(
                        data, partition_indices, order_by_cols, col_name, is_dense
                    )
            else:
                # No partitioning, apply to entire dataset
                self._apply_rank_to_partition(
                    data, list(range(len(data))), order_by_cols, col_name, is_dense
                )
        else:
            # No window spec, assign ranks based on original order
            for i in range(len(data)):
                data[i][col_name] = i + 1

    def _apply_rank_to_partition(
        self,
        data: List[Dict[str, Any]],
        indices: List[int],
        order_by_cols: List[Any],
        col_name: str,
        is_dense: bool,
    ) -> None:
        """Apply rank or dense_rank to a specific partition."""
        if not order_by_cols:
            # No order by, assign ranks based on original order
            for i, idx in enumerate(indices):
                data[idx][col_name] = i + 1
            return

        # Sort partition by order by columns using the corrected ordering logic
        sorted_indices = self._apply_ordering_to_indices(data, indices, order_by_cols)

        # Assign ranks in sorted order
        if is_dense:
            # Dense rank: consecutive ranks without gaps
            current_rank = 1
            previous_values = None

            for i, idx in enumerate(sorted_indices):
                row = data[idx]
                current_values = []
                for col in order_by_cols:
                    # Handle MockColumnOperation objects (like col("salary").desc())
                    if hasattr(col, "column") and hasattr(col.column, "name"):
                        order_col_name = col.column.name
                    elif hasattr(col, "name"):
                        order_col_name = col.name
                    else:
                        order_col_name = str(col)
                    value = row.get(order_col_name)
                    current_values.append(value)

                if previous_values is not None and current_values != previous_values:
                    current_rank += 1

                data[idx][col_name] = current_rank
                previous_values = current_values
        else:
            # Regular rank: ranks with gaps for ties
            current_rank = 1

            for i, idx in enumerate(sorted_indices):
                if i > 0:
                    prev_idx = sorted_indices[i - 1]
                    # Check if current and previous rows have different values
                    row = data[idx]
                    prev_row = data[prev_idx]

                    current_values = []
                    prev_values = []
                    for col in order_by_cols:
                        # Handle MockColumnOperation objects (like col("salary").desc())
                        if hasattr(col, "column") and hasattr(col.column, "name"):
                            order_col_name = col.column.name
                        elif hasattr(col, "name"):
                            order_col_name = col.name
                        else:
                            order_col_name = str(col)
                        current_values.append(row.get(order_col_name))
                        prev_values.append(prev_row.get(order_col_name))

                    if current_values != prev_values:
                        current_rank = i + 1
                else:
                    current_rank = 1

                data[idx][col_name] = current_rank

    def _evaluate_aggregate_window_functions(
        self, data: List[Dict[str, Any]], window_func: Any, col_name: str
    ) -> None:
        """Evaluate aggregate window functions like avg, sum, count, etc."""
        if not window_func.column_name and window_func.function_name not in ["count"]:
            # No column specified for functions that need it
            for row in data:
                row[col_name] = None
            return

        # Handle window specification if present
        if hasattr(window_func, "window_spec") and window_func.window_spec:
            window_spec = window_func.window_spec
            partition_by_cols = getattr(window_spec, "_partition_by", [])
            order_by_cols = getattr(window_spec, "_order_by", [])

            if partition_by_cols:
                # Handle partitioning
                partition_groups: Dict[Any, List[int]] = {}
                for i, row in enumerate(data):
                    partition_key = tuple(
                        row.get(col.name) if hasattr(col, "name") else row.get(str(col))
                        for col in partition_by_cols
                    )
                    if partition_key not in partition_groups:
                        partition_groups[partition_key] = []
                    partition_groups[partition_key].append(i)

                # Process each partition
                for partition_indices in partition_groups.values():
                    # Apply ordering to partition indices
                    ordered_indices = self._apply_ordering_to_indices(
                        data, partition_indices, order_by_cols
                    )
                    self._apply_aggregate_to_partition(
                        data, ordered_indices, window_func, col_name
                    )
            else:
                # No partitioning, apply to entire dataset with ordering
                all_indices = list(range(len(data)))
                ordered_indices = self._apply_ordering_to_indices(
                    data, all_indices, order_by_cols
                )
                self._apply_aggregate_to_partition(
                    data, ordered_indices, window_func, col_name
                )
        else:
            # No window spec, apply to entire dataset
            all_indices = list(range(len(data)))
            self._apply_aggregate_to_partition(data, all_indices, window_func, col_name)

    def _apply_aggregate_to_partition(
        self,
        data: List[Dict[str, Any]],
        indices: List[int],
        window_func: Any,
        col_name: str,
    ) -> None:
        """Apply aggregate function to a specific partition."""
        if not indices:
            return

        source_col = window_func.column_name
        func_name = window_func.function_name

        # Get window boundaries if specified
        rows_between = (
            getattr(window_func.window_spec, "_rows_between", None)
            if hasattr(window_func, "window_spec") and window_func.window_spec
            else None
        )

        for i, idx in enumerate(indices):
            # Determine the window for this row
            if rows_between:
                start_offset, end_offset = rows_between
                window_start = max(0, i + start_offset)
                window_end = min(len(indices), i + end_offset + 1)
            else:
                # Default: all rows up to current row
                window_start = 0
                window_end = i + 1

            # Get values in the window
            window_values = []
            for j in range(window_start, window_end):
                if j < len(indices):
                    row_idx = indices[j]
                    if source_col:
                        value = data[row_idx].get(source_col)
                        if value is not None:
                            window_values.append(value)
                    else:
                        # For count(*) - count all rows
                        window_values.append(1)

            # Apply aggregate function
            if func_name == "avg":
                data[idx][col_name] = (
                    sum(window_values) / len(window_values) if window_values else None
                )
            elif func_name == "sum":
                data[idx][col_name] = sum(window_values) if window_values else None
            elif func_name == "count":
                data[idx][col_name] = len(window_values)
            elif func_name == "countDistinct":
                data[idx][col_name] = len(set(window_values))
            elif func_name == "max":
                data[idx][col_name] = max(window_values) if window_values else None
            elif func_name == "min":
                data[idx][col_name] = min(window_values) if window_values else None



"""
Performance simulation framework for Mock Spark.

This module provides performance simulation capabilities that allow testing
of performance-related scenarios by simulating delays, memory constraints,
and other performance characteristics without requiring actual large datasets.

Key Features:
    - Configurable slowdown simulation
    - Memory limit simulation
    - Performance monitoring and metrics
    - Realistic performance testing scenarios
    - Easy integration with existing test frameworks

Example:
    >>> from mock_spark import MockSparkSession
    >>> from mock_spark.performance_simulation import MockPerformanceSimulator
    >>>
    >>> spark = MockSparkSession("test")
    >>> perf_sim = MockPerformanceSimulator(spark)
    >>>
    >>> # Simulate slow operations
    >>> perf_sim.set_slowdown(2.0)  # 2x slower
    >>>
    >>> # Simulate memory constraints
    >>> perf_sim.set_memory_limit(1000)  # 1000 rows max
"""

import time
from typing import Any, Callable, Dict, Optional
from contextlib import contextmanager


class MockPerformanceSimulator:
    """Performance simulation framework for Mock Spark operations.

    Provides configurable performance simulation including slowdown factors,
    memory limits, and performance monitoring for realistic testing scenarios.

    Attributes:
        spark_session: The MockSparkSession instance to simulate performance on.
        slowdown_factor: Multiplier for operation delays (1.0 = normal speed).
        memory_limit: Maximum number of rows allowed in operations.
        performance_metrics: Dictionary storing performance statistics.

    Example:
        >>> perf_sim = MockPerformanceSimulator(spark)
        >>> perf_sim.set_slowdown(2.0)
        >>> perf_sim.set_memory_limit(1000)
    """

    def __init__(self, spark_session: Any) -> None:
        """Initialize MockPerformanceSimulator.

        Args:
            spark_session: MockSparkSession instance to simulate performance on.
        """
        self.spark_session = spark_session
        self.slowdown_factor = 1.0
        self.memory_limit: Optional[int] = None
        self.performance_metrics = {
            "total_operations": 0,
            "total_time": 0.0,
            "memory_usage": 0,
            "slowdown_applied": 0,
            "memory_limits_hit": 0,
        }
        self._original_methods: Dict[str, Any] = {}

    def set_slowdown(self, factor: float) -> None:
        """Set slowdown factor for operations.

        Args:
            factor: Slowdown multiplier (1.0 = normal speed, 2.0 = 2x slower).
        """
        if factor < 0:
            raise ValueError("Slowdown factor must be non-negative")
        self.slowdown_factor = factor

    def set_memory_limit(self, max_rows: int) -> None:
        """Set memory limit for operations.

        Args:
            max_rows: Maximum number of rows allowed in operations.
        """
        if max_rows < 0:
            raise ValueError("Memory limit must be non-negative")
        self.memory_limit = max_rows

    def check_memory_usage(self, data_size: int) -> None:
        """Check if operation exceeds memory limit.

        Args:
            data_size: Number of rows in the operation.

        Raises:
            PySparkRuntimeError: If memory limit is exceeded.
        """
        if self.memory_limit is not None and data_size > self.memory_limit:
            from .errors import PySparkRuntimeError

            self.performance_metrics["memory_limits_hit"] += 1
            raise PySparkRuntimeError(
                f"Out of memory: {data_size} rows exceeds limit of {self.memory_limit}"
            )

    def simulate_slow_operation(
        self, operation: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Any:
        """Simulate a slow operation with configurable delay.

        Args:
            operation: Function to execute with slowdown.
            *args: Positional arguments for the operation.
            **kwargs: Keyword arguments for the operation.

        Returns:
            Result of the operation.
        """
        start_time = time.time()

        # Apply slowdown
        if self.slowdown_factor > 1.0:
            delay = 0.001 * (self.slowdown_factor - 1.0)  # 1ms per slowdown unit
            time.sleep(delay)
            self.performance_metrics["slowdown_applied"] += 1

        # Execute operation
        result = operation(*args, **kwargs)

        # Record metrics
        end_time = time.time()
        self.performance_metrics["total_operations"] += 1
        self.performance_metrics["total_time"] += end_time - start_time

        return result

    def get_performance_metrics(self) -> Dict[str, Any]:
        """Get current performance metrics.

        Returns:
            Dictionary containing performance statistics.
        """
        metrics = self.performance_metrics.copy()
        if metrics["total_operations"] > 0:
            metrics["average_time_per_operation"] = (
                metrics["total_time"] / metrics["total_operations"]
            )
        else:
            metrics["average_time_per_operation"] = 0.0
        return metrics

    def reset_metrics(self) -> None:
        """Reset performance metrics to zero."""
        self.performance_metrics = {
            "total_operations": 0,
            "total_time": 0.0,
            "memory_usage": 0,
            "slowdown_applied": 0,
            "memory_limits_hit": 0,
        }

    def enable_performance_simulation(self) -> None:
        """Enable performance simulation by wrapping methods."""
        # Store original methods
        self._original_methods = {
            "createDataFrame": self.spark_session.createDataFrame,
            "table": self.spark_session.table,
            "sql": self.spark_session.sql,
        }

        # Wrap methods with performance simulation
        self.spark_session.createDataFrame = self._wrap_method("createDataFrame")
        self.spark_session.table = self._wrap_method("table")
        self.spark_session.sql = self._wrap_method("sql")

    def disable_performance_simulation(self) -> None:
        """Disable performance simulation by restoring original methods."""
        for method_name, original_method in self._original_methods.items():
            setattr(self.spark_session, method_name, original_method)
        self._original_methods.clear()

    def _wrap_method(self, method_name: str) -> Callable[..., Any]:
        """Wrap a method with performance simulation logic."""
        original_method = getattr(self.spark_session, method_name)

        def wrapped_method(*args: Any, **kwargs: Any) -> Any:
            # Check memory limits for createDataFrame
            if method_name == "createDataFrame" and args:
                data = args[0]
                if isinstance(data, list):
                    self.check_memory_usage(len(data))

            # Simulate slow operation
            return self.simulate_slow_operation(original_method, *args, **kwargs)

        return wrapped_method


class MockPerformanceSimulatorBuilder:
    """Builder pattern for creating common performance simulation scenarios.

    Provides convenient methods for setting up common performance testing
    scenarios without requiring manual configuration.

    Example:
        >>> builder = MockPerformanceSimulatorBuilder(spark)
        >>> perf_sim = (builder
        ...     .slowdown(2.0)
        ...     .memory_limit(1000)
        ...     .enable_monitoring()
        ...     .build())
    """

    def __init__(self, spark_session: Any) -> None:
        """Initialize MockPerformanceSimulatorBuilder.

        Args:
            spark_session: MockSparkSession instance to build performance simulation for.
        """
        self.spark_session = spark_session
        self.perf_sim = MockPerformanceSimulator(spark_session)

    def slowdown(self, factor: float) -> "MockPerformanceSimulatorBuilder":
        """Set slowdown factor.

        Args:
            factor: Slowdown multiplier.
        """
        self.perf_sim.set_slowdown(factor)
        return self

    def memory_limit(self, max_rows: int) -> "MockPerformanceSimulatorBuilder":
        """Set memory limit.

        Args:
            max_rows: Maximum number of rows allowed.
        """
        self.perf_sim.set_memory_limit(max_rows)
        return self

    def enable_monitoring(self) -> "MockPerformanceSimulatorBuilder":
        """Enable performance monitoring."""
        self.perf_sim.enable_performance_simulation()
        return self

    def build(self) -> MockPerformanceSimulator:
        """Build the performance simulator with all configured settings.

        Returns:
            MockPerformanceSimulator instance with all settings applied.
        """
        return self.perf_sim


@contextmanager
def performance_simulation(
    spark_session: Any, slowdown_factor: float = 1.0, memory_limit: Optional[int] = None
) -> Any:
    """Context manager for temporary performance simulation.

    Args:
        spark_session: MockSparkSession instance.
        slowdown_factor: Slowdown multiplier for operations.
        memory_limit: Maximum number of rows allowed.

    Example:
        >>> with performance_simulation(spark, slowdown_factor=2.0, memory_limit=1000):
        ...     df = spark.createDataFrame(data)
    """
    perf_sim = MockPerformanceSimulator(spark_session)
    perf_sim.set_slowdown(slowdown_factor)
    if memory_limit:
        perf_sim.set_memory_limit(memory_limit)

    try:
        perf_sim.enable_performance_simulation()
        yield perf_sim
    finally:
        perf_sim.disable_performance_simulation()


# Convenience functions for common performance scenarios
def create_slow_simulator(
    spark_session: Any, slowdown_factor: float = 2.0
) -> MockPerformanceSimulator:
    """Create a simulator that slows down operations.

    Args:
        spark_session: MockSparkSession instance.
        slowdown_factor: Slowdown multiplier.

    Returns:
        MockPerformanceSimulator configured for slow operations.
    """
    return (
        MockPerformanceSimulatorBuilder(spark_session).slowdown(slowdown_factor).build()
    )


def create_memory_limited_simulator(
    spark_session: Any, memory_limit: int = 1000
) -> MockPerformanceSimulator:
    """Create a simulator that limits memory usage.

    Args:
        spark_session: MockSparkSession instance.
        memory_limit: Maximum number of rows allowed.

    Returns:
        MockPerformanceSimulator configured for memory limits.
    """
    return (
        MockPerformanceSimulatorBuilder(spark_session)
        .memory_limit(memory_limit)
        .build()
    )


def create_high_performance_simulator(spark_session: Any) -> MockPerformanceSimulator:
    """Create a simulator optimized for high performance.

    Args:
        spark_session: MockSparkSession instance.

    Returns:
        MockPerformanceSimulator configured for high performance.
    """
    return MockPerformanceSimulatorBuilder(spark_session).slowdown(0.5).build()



"""
Adapter to convert spark-ddl-parser output to MockStructType.

This module provides an adapter layer between the standalone spark-ddl-parser
package and mock-spark's internal type system.
"""

from spark_ddl_parser import parse_ddl_schema as parse_ddl
from spark_ddl_parser.types import (
    StructType,
    StructField,
    SimpleType,
    DecimalType,
    ArrayType,
    MapType,
    DataType,
)

from ..spark_types import (
    MockStructType,
    MockStructField,
    MockDataType,
    StringType,
    IntegerType,
    LongType,
    DoubleType,
    BooleanType,
    DateType,
    TimestampType,
    BinaryType,
    FloatType,
    ShortType,
    ByteType,
    DecimalType as MockDecimalType,
    ArrayType as MockArrayType,
    MapType as MockMapType,
)


def parse_ddl_schema(ddl_string: str) -> MockStructType:
    """Parse DDL and convert to MockStructType.

    Args:
        ddl_string: DDL schema string (e.g., "id long, name string")

    Returns:
        MockStructType with parsed fields

    Raises:
        ValueError: If DDL string is invalid
    """
    parsed = parse_ddl(ddl_string)
    return _convert_struct_type(parsed)


def _convert_struct_type(struct: StructType) -> MockStructType:
    """Convert StructType to MockStructType.

    Args:
        struct: Parsed StructType from spark-ddl-parser

    Returns:
        MockStructType
    """
    fields = [_convert_field(f) for f in struct.fields]
    return MockStructType(fields)


def _convert_field(field: StructField) -> MockStructField:
    """Convert StructField to MockStructField.

    Args:
        field: Parsed StructField from spark-ddl-parser

    Returns:
        MockStructField
    """
    data_type = _convert_data_type(field.data_type)
    return MockStructField(name=field.name, dataType=data_type, nullable=field.nullable)


def _convert_data_type(data_type: DataType) -> MockDataType:
    """Convert DataType to MockDataType.

    Args:
        data_type: Parsed DataType from spark-ddl-parser

    Returns:
        MockDataType
    """
    if isinstance(data_type, SimpleType):
        return _convert_simple_type(data_type)
    elif isinstance(data_type, DecimalType):
        return MockDecimalType(precision=data_type.precision, scale=data_type.scale)
    elif isinstance(data_type, ArrayType):
        element_type = _convert_data_type(data_type.element_type)
        return MockArrayType(element_type)
    elif isinstance(data_type, MapType):
        key_type = _convert_data_type(data_type.key_type)
        value_type = _convert_data_type(data_type.value_type)
        return MockMapType(key_type, value_type)
    elif isinstance(data_type, StructType):
        return _convert_struct_type(data_type)
    else:
        # Default to string for unknown types
        return StringType()


def _convert_simple_type(simple_type: SimpleType) -> MockDataType:
    """Convert SimpleType to appropriate MockDataType.

    Args:
        simple_type: SimpleType from spark-ddl-parser

    Returns:
        MockDataType instance
    """
    type_mapping = {
        "string": StringType,
        "integer": IntegerType,
        "long": LongType,
        "double": DoubleType,
        "float": FloatType,
        "boolean": BooleanType,
        "date": DateType,
        "timestamp": TimestampType,
        "binary": BinaryType,
        "short": ShortType,
        "byte": ByteType,
    }

    type_class = type_mapping.get(simple_type.type_name, StringType)
    return type_class()



"""
DuckDB-based materialization for lazy evaluation.

This module provides DuckDB-specific optimizations on top of the generic SQLAlchemy materializer.
"""

from .query_executor import SQLAlchemyMaterializer


class DuckDBMaterializer(SQLAlchemyMaterializer):
    """Materializes lazy DataFrames using DuckDB with SQLAlchemy."""

    def __init__(self, max_memory: str = "1GB", allow_disk_spillover: bool = False):
        """Initialize DuckDB materializer.

        Args:
            max_memory: Maximum memory for DuckDB to use (e.g., '1GB', '4GB', '8GB').
            allow_disk_spillover: If True, allows DuckDB to spill to disk when memory is full.
        """
        # Initialize with DuckDB engine
        super().__init__(engine_url="duckdb:///:memory:")

        self._temp_dir = None

        # Configure DuckDB-specific settings using raw connection
        try:
            raw_conn = self.engine.raw_connection()
            raw_conn.execute(f"SET max_memory='{max_memory}'")

            if allow_disk_spillover:
                # Create unique temp directory for this materializer
                import tempfile
                import uuid

                self._temp_dir = tempfile.mkdtemp(
                    prefix=f"duckdb_mat_{uuid.uuid4().hex[:8]}_"
                )
                raw_conn.execute(f"SET temp_directory='{self._temp_dir}'")
            else:
                # Disable disk spillover for test isolation
                raw_conn.execute("SET temp_directory=''")
        except:  # noqa: E722
            pass  # Ignore if settings not supported

    def close(self) -> None:
        """Close the DuckDB connection and clean up temp directory."""
        # Clean up unique temp directory if it exists
        if self._temp_dir:
            try:
                import os
                import shutil

                if os.path.exists(self._temp_dir):
                    shutil.rmtree(self._temp_dir, ignore_errors=True)
                self._temp_dir = None
            except:  # noqa: E722
                pass  # Ignore cleanup errors

        # Call parent close to dispose engine
        super().close()

    def __del__(self) -> None:
        """Cleanup on deletion to prevent resource leaks."""
        try:
            self.close()
        except:  # noqa: E722
            pass



"""
Runtime exception classes for Mock Spark.

This module provides exception classes for runtime-related errors,
including system errors, attribute errors, and runtime failures.
"""

from typing import Any, Optional
from .base import MockSparkException


class PySparkRuntimeError(MockSparkException):
    """PySpark-compatible RuntimeError.

    Raised when a runtime error occurs during execution
    of Spark operations.

    Args:
        message: Error message describing the runtime error.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise PySparkRuntimeError("Runtime error occurred")
    """

    def __init__(self, message: str, stackTrace: Optional[Any] = None):
        super().__init__(message, stackTrace)


class PySparkAttributeError(MockSparkException):
    """PySpark-compatible AttributeError.

    Raised when an attribute reference or assignment fails
    on a Spark object.

    Args:
        object_name: Name of the object that lacks the attribute.
        attribute_name: Name of the missing attribute.
        message: Optional custom error message.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise PySparkAttributeError("DataFrame", "unknown_method")
    """

    def __init__(
        self,
        object_name: str,
        attribute_name: str,
        message: Optional[str] = None,
        stackTrace: Optional[Any] = None,
    ):
        if message is None:
            message = f"'{object_name}' object has no attribute '{attribute_name}'"
        super().__init__(message, stackTrace)
        self.object_name = object_name
        self.attribute_name = attribute_name


class SystemException(PySparkRuntimeError):
    """Exception raised for system-level errors.

    Raised when there are system-level issues that prevent
    normal operation of Spark components.

    Args:
        message: Error message describing the system error.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise SystemException("System resource unavailable")
    """

    def __init__(self, message: str, stackTrace: Optional[Any] = None):
        super().__init__(message, stackTrace)


class ConfigurationException(PySparkRuntimeError):
    """Exception raised for configuration errors.

    Raised when there are issues with configuration
    management or invalid configuration values.

    Args:
        config_key: Configuration key that caused the error.
        config_value: Configuration value that caused the error.
        message: Optional custom error message.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise ConfigurationException("spark.memory", "invalid_value")
    """

    def __init__(
        self,
        config_key: str,
        config_value: Any,
        message: Optional[str] = None,
        stackTrace: Optional[Any] = None,
    ):
        if message is None:
            message = (
                f"Invalid configuration value '{config_value}' for key '{config_key}'"
            )
        super().__init__(message, stackTrace)
        self.config_key = config_key
        self.config_value = config_value


class InitializationException(PySparkRuntimeError):
    """Exception raised for initialization errors.

    Raised when there are issues during initialization
    of Spark components or sessions.

    Args:
        component: Name of the component that failed to initialize.
        message: Optional custom error message.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise InitializationException("SparkSession")
    """

    def __init__(
        self,
        component: str,
        message: Optional[str] = None,
        stackTrace: Optional[Any] = None,
    ):
        if message is None:
            message = f"Failed to initialize {component}"
        super().__init__(message, stackTrace)
        self.component = component


class ShutdownException(PySparkRuntimeError):
    """Exception raised for shutdown errors.

    Raised when there are issues during shutdown
    of Spark components or sessions.

    Args:
        component: Name of the component that failed to shutdown.
        message: Optional custom error message.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise ShutdownException("SparkSession")
    """

    def __init__(
        self,
        component: str,
        message: Optional[str] = None,
        stackTrace: Optional[Any] = None,
    ):
        if message is None:
            message = f"Failed to shutdown {component}"
        super().__init__(message, stackTrace)
        self.component = component


class TimeoutException(PySparkRuntimeError):
    """Exception raised for timeout errors.

    Raised when operations exceed their timeout limits
    or when waiting for resources times out.

    Args:
        operation: Name of the operation that timed out.
        timeout_seconds: Timeout value in seconds.
        message: Optional custom error message.
        stackTrace: Optional stack trace information.

    Example:
        >>> raise TimeoutException("query_execution", 30)
    """

    def __init__(
        self,
        operation: str,
        timeout_seconds: int,
        message: Optional[str] = None,
        stackTrace: Optional[Any] = None,
    ):
        if message is None:
            message = (
                f"Operation '{operation}' timed out after {timeout_seconds} seconds"
            )
        super().__init__(message, stackTrace)
        self.operation = operation
        self.timeout_seconds = timeout_seconds



"""
Query Optimizer for Mock Spark.

This module provides query optimization functionality for Mock Spark,
analyzing and optimizing SQL queries for better performance.
It includes rule-based optimization, cost estimation, and query plan generation.

Key Features:
    - Query plan optimization
    - Cost-based optimization rules
    - Join order optimization
    - Predicate pushdown
    - Projection elimination

Example:
    >>> from mock_spark.session.sql import MockQueryOptimizer
    >>> optimizer = MockQueryOptimizer()
    >>> optimized_ast = optimizer.optimize(ast)
"""

from typing import Any, Dict, List, Optional
from .parser import MockSQLAST


class MockQueryPlan:
    """Query execution plan for Mock Spark."""

    def __init__(
        self,
        plan_type: str,
        children: Optional[List["MockQueryPlan"]] = None,
        properties: Optional[Dict[str, Any]] = None,
    ):
        """Initialize query plan.

        Args:
            plan_type: Type of query plan node.
            children: Child query plan nodes.
            properties: Plan node properties.
        """
        self.plan_type = plan_type
        self.children = children or []
        self.properties = properties or {}

    def __str__(self) -> str:
        """String representation."""
        return f"MockQueryPlan(type='{self.plan_type}', children={len(self.children)})"

    def __repr__(self) -> str:
        """Representation."""
        return self.__str__()


class MockQueryOptimizer:
    """Query Optimizer for Mock Spark.

    Provides query optimization functionality that analyzes SQL queries
    and generates optimized execution plans. Includes rule-based optimization,
    cost estimation, and various optimization techniques.

    Example:
        >>> optimizer = MockQueryOptimizer()
        >>> optimized_ast = optimizer.optimize(ast)
    """

    def __init__(self) -> None:
        """Initialize MockQueryOptimizer."""
        self._optimization_rules = [
            self._rule_predicate_pushdown,
            self._rule_projection_elimination,
            self._rule_join_reordering,
            self._rule_constant_folding,
            self._rule_redundant_operations,
        ]

    def optimize(self, ast: MockSQLAST) -> MockSQLAST:
        """Optimize SQL AST.

        Args:
            ast: Parsed SQL AST.

        Returns:
            Optimized SQL AST.
        """
        if ast.query_type != "SELECT":
            return ast  # Only optimize SELECT queries for now

        optimized_components = ast.components.copy()

        # Apply optimization rules
        for rule in self._optimization_rules:
            optimized_components = rule(optimized_components)

        # Create optimized AST
        optimized_components["original_query"] = ast.components.get(
            "original_query", ""
        )
        optimized_ast = MockSQLAST(ast.query_type, optimized_components)

        return optimized_ast

    def generate_execution_plan(self, ast: MockSQLAST) -> MockQueryPlan:
        """Generate execution plan for SQL AST.

        Args:
            ast: Parsed SQL AST.

        Returns:
            Query execution plan.
        """
        if ast.query_type == "SELECT":
            return self._generate_select_plan(ast)
        else:
            return MockQueryPlan("UNKNOWN", properties={"query_type": ast.query_type})

    def _rule_predicate_pushdown(self, components: Dict[str, Any]) -> Dict[str, Any]:
        """Apply predicate pushdown optimization.

        Args:
            components: Query components.

        Returns:
            Optimized components.
        """
        # Mock implementation - in real optimizer this would push WHERE conditions
        # down to the data source level
        return components

    def _rule_projection_elimination(
        self, components: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Apply projection elimination optimization.

        Args:
            components: Query components.

        Returns:
            Optimized components.
        """
        # Mock implementation - in real optimizer this would eliminate
        # unnecessary columns early in the pipeline
        return components

    def _rule_join_reordering(self, components: Dict[str, Any]) -> Dict[str, Any]:
        """Apply join reordering optimization.

        Args:
            components: Query components.

        Returns:
            Optimized components.
        """
        # Mock implementation - in real optimizer this would reorder joins
        # for better performance
        return components

    def _rule_constant_folding(self, components: Dict[str, Any]) -> Dict[str, Any]:
        """Apply constant folding optimization.

        Args:
            components: Query components.

        Returns:
            Optimized components.
        """
        # Mock implementation - in real optimizer this would evaluate
        # constant expressions at compile time
        return components

    def _rule_redundant_operations(self, components: Dict[str, Any]) -> Dict[str, Any]:
        """Apply redundant operations elimination.

        Args:
            components: Query components.

        Returns:
            Optimized components.
        """
        # Mock implementation - in real optimizer this would eliminate
        # redundant operations
        return components

    def _generate_select_plan(self, ast: MockSQLAST) -> MockQueryPlan:
        """Generate execution plan for SELECT query.

        Args:
            ast: SELECT query AST.

        Returns:
            SELECT query execution plan.
        """
        components = ast.components

        # Create a simple execution plan
        plan = MockQueryPlan(
            "SELECT",
            properties={
                "columns": components.get("select_columns", ["*"]),
                "tables": components.get("from_tables", []),
                "conditions": components.get("where_conditions", []),
                "group_by": components.get("group_by_columns", []),
                "order_by": components.get("order_by_columns", []),
                "limit": components.get("limit_value"),
            },
        )

        # Add child plans for different operations
        if components.get("where_conditions"):
            filter_plan = MockQueryPlan(
                "FILTER",
                properties={"conditions": components.get("where_conditions", [])},
            )
            plan.children.append(filter_plan)

        if components.get("group_by_columns"):
            group_plan = MockQueryPlan(
                "GROUP_BY",
                properties={"columns": components.get("group_by_columns", [])},
            )
            plan.children.append(group_plan)

        if components.get("order_by_columns"):
            sort_plan = MockQueryPlan(
                "SORT", properties={"columns": components.get("order_by_columns", [])}
            )
            plan.children.append(sort_plan)

        if components.get("limit_value"):
            limit_plan = MockQueryPlan(
                "LIMIT", properties={"limit": components.get("limit_value")}
            )
            plan.children.append(limit_plan)

        return plan

    def estimate_cost(self, plan: MockQueryPlan) -> float:
        """Estimate cost of execution plan.

        Args:
            plan: Query execution plan.

        Returns:
            Estimated cost.
        """
        # Mock implementation - in real optimizer this would calculate
        # actual cost based on statistics and heuristics
        base_cost = 1.0

        # Add cost for each operation
        for child in plan.children:
            if child.plan_type == "FILTER":
                base_cost += 0.1
            elif child.plan_type == "GROUP_BY":
                base_cost += 0.5
            elif child.plan_type == "SORT":
                base_cost += 0.3
            elif child.plan_type == "LIMIT":
                base_cost += 0.05

        return base_cost



"""
Error handling for Mock Spark.

DEPRECATED: This module re-exports exceptions from core.exceptions for backward compatibility.
New code should import directly from mock_spark.core.exceptions.

This module provides comprehensive error handling that matches PySpark's
exception hierarchy for maximum compatibility.

Example:
    >>> from mock_spark.errors import AnalysisException
    >>> raise AnalysisException("Column 'unknown' does not exist")
"""

# Re-export from core.exceptions for backward compatibility
from .core.exceptions.base import MockException, MockSparkException
from .core.exceptions.analysis import (
    AnalysisException,
    ParseException,
    SchemaException,
    ColumnNotFoundException,
    TableNotFoundException,
    DatabaseNotFoundException,
    TypeMismatchException,
)
from .core.exceptions.execution import (
    QueryExecutionException,
    SparkUpgradeException,
    StreamingQueryException,
    TempTableAlreadyExistsException,
    UnsupportedOperationException,
    ResourceException,
    MemoryException,
)
from .core.exceptions.validation import (
    IllegalArgumentException,
    PySparkValueError,
    PySparkTypeError,
    ValidationException,
)
from .core.exceptions.runtime import (
    PySparkRuntimeError,
    PySparkAttributeError,
    ConfigurationException,
)

# Alias for PySparkException (same as MockSparkException)
PySparkException = MockSparkException

# Mark as available
PYSPARK_AVAILABLE = True


# Helper functions for common error scenarios
def raise_table_not_found(table_name: str) -> None:
    """Raise table not found error."""
    raise TableNotFoundException(f"Table or view not found: {table_name}")


def raise_column_not_found(column_name: str) -> None:
    """Raise column not found error."""
    raise ColumnNotFoundException(f"Column '{column_name}' does not exist")


def raise_schema_not_found(schema_name: str) -> None:
    """Raise schema not found error."""
    raise DatabaseNotFoundException(f"Database '{schema_name}' not found")


def raise_invalid_argument(param_name: str, value: str, expected: str) -> None:
    """Raise invalid argument error."""
    raise IllegalArgumentException(
        f"Invalid value for parameter '{param_name}': {value}. Expected: {expected}"
    )


def raise_unsupported_operation(operation: str) -> None:
    """Raise unsupported operation error."""
    raise UnsupportedOperationException(
        f"Operation '{operation}' is not supported in mock mode"
    )


def raise_parse_error(sql: str, error: str) -> None:
    """Raise parse error."""
    raise ParseException(f"Error parsing SQL: {sql}. {error}")


def raise_query_execution_error(error: str) -> None:
    """Raise query execution error."""
    raise QueryExecutionException(f"Query execution failed: {error}")


def raise_type_error(expected_type: str, actual_type: str) -> None:
    """Raise type error."""
    raise PySparkTypeError(f"Expected {expected_type}, got {actual_type}")


def raise_value_error(message: str) -> None:
    """Raise value error."""
    raise PySparkValueError(message)


def raise_runtime_error(message: str) -> None:
    """Raise runtime error."""
    raise PySparkRuntimeError(message)


# Export commonly used exceptions
__all__ = [
    # Base exceptions
    "MockException",
    "MockSparkException",
    "PySparkException",
    # Analysis exceptions
    "AnalysisException",
    "ParseException",
    "SchemaException",
    "ColumnNotFoundException",
    "TableNotFoundException",
    "DatabaseNotFoundException",
    "TypeMismatchException",
    # Execution exceptions
    "QueryExecutionException",
    "SparkUpgradeException",
    "StreamingQueryException",
    "TempTableAlreadyExistsException",
    "UnsupportedOperationException",
    "ResourceException",
    "MemoryException",
    # Validation exceptions
    "IllegalArgumentException",
    "PySparkValueError",
    "PySparkTypeError",
    "ValidationException",
    # Runtime exceptions
    "PySparkRuntimeError",
    "PySparkAttributeError",
    "ConfigurationException",
    # Helper functions
    "raise_table_not_found",
    "raise_column_not_found",
    "raise_schema_not_found",
    "raise_invalid_argument",
    "raise_unsupported_operation",
    "raise_parse_error",
    "raise_query_execution_error",
    "raise_type_error",
    "raise_value_error",
    "raise_runtime_error",
    # Flags
    "PYSPARK_AVAILABLE",
]



"""
Type protocols for Mock Spark.

This module defines structural typing protocols (PEP 544) for better
type safety and clearer contracts without tight coupling.
"""

from typing import Protocol, Any, runtime_checkable


@runtime_checkable
class ColumnLike(Protocol):
    """Protocol for column-like objects."""

    @property
    def name(self) -> str:
        """Column name."""
        ...


@runtime_checkable
class OperationLike(Protocol):
    """Protocol for column operation objects."""

    @property
    def column(self) -> Any:
        """The column being operated on."""
        ...

    @property
    def operation(self) -> str:
        """The operation type."""
        ...

    @property
    def value(self) -> Any:
        """The operation value/operand."""
        ...

    @property
    def name(self) -> str:
        """Operation name."""
        ...


@runtime_checkable
class LiteralLike(Protocol):
    """Protocol for literal value objects."""

    @property
    def value(self) -> Any:
        """The literal value."""
        ...

    @property
    def name(self) -> str:
        """Literal name."""
        ...


@runtime_checkable
class CaseWhenLike(Protocol):
    """Protocol for CASE WHEN expression objects."""

    @property
    def conditions(self) -> Any:
        """List of (condition, value) tuples."""
        ...

    @property
    def default_value(self) -> Any:
        """Default value for ELSE clause."""
        ...


@runtime_checkable
class DataFrameLike(Protocol):
    """Protocol for DataFrame-like objects."""

    @property
    def data(self) -> Any:
        """DataFrame data."""
        ...

    @property
    def schema(self) -> Any:
        """DataFrame schema."""
        ...

    def collect(self) -> Any:
        """Collect DataFrame rows."""
        ...


@runtime_checkable
class SchemaLike(Protocol):
    """Protocol for schema-like objects."""

    @property
    def fields(self) -> Any:
        """Schema fields."""
        ...

    def fieldNames(self) -> Any:
        """Get field names."""
        ...


# Type aliases for common unions
ColumnExpression = Any  # Can be ColumnLike, OperationLike, LiteralLike, or str
AggregateExpression = Any  # Can be string name or column operation
WindowExpression = Any  # Window function with spec



"""
Expression functions for Mock Spark.

This module provides the F namespace functions and expression utilities
for creating column expressions and transformations.
"""

from typing import Any, Union, List, TYPE_CHECKING
from .column import MockColumn, MockColumnOperation
from .literals import MockLiteral

if TYPE_CHECKING:
    from ..conditional import MockCaseWhen


class ExpressionFunctions:
    """Expression functions for creating column expressions."""

    @staticmethod
    def col(name: str) -> MockColumn:
        """Create a column reference.

        Delegates to canonical MockColumn constructor.

        Args:
            name: Column name.

        Returns:
            MockColumn instance.
        """
        return MockColumn(name)

    @staticmethod
    def lit(value: Any) -> MockLiteral:
        """Create a literal value.

        Delegates to canonical MockLiteral constructor.

        Args:
            value: Literal value.

        Returns:
            MockLiteral instance.
        """
        return MockLiteral(value)

    @staticmethod
    def when(condition: MockColumnOperation, value: Any) -> "MockCaseWhen":
        """Start a CASE WHEN expression.

        Delegates to canonical MockCaseWhen constructor.

        Args:
            condition: Condition to evaluate.
            value: Value if condition is true.

        Returns:
            MockCaseWhen instance.
        """
        from ..conditional import MockCaseWhen

        return MockCaseWhen(None, condition, value)

    @staticmethod
    def coalesce(
        *columns: Union[MockColumn, MockColumnOperation, str],
    ) -> MockColumnOperation:
        """Return the first non-null value from a list of columns.

        Args:
            *columns: Columns to check for non-null values.

        Returns:
            MockColumnOperation for coalesce.
        """
        col_refs: List[Union[MockColumn, MockColumnOperation]] = []
        for col in columns:
            if isinstance(col, str):
                col_refs.append(MockColumn(col))
            else:
                col_refs.append(col)

        return MockColumnOperation(None, "coalesce", col_refs)

    @staticmethod
    def isnull(column: Union[MockColumn, str]) -> MockColumnOperation:
        """Check if column value is null.

        Args:
            column: Column to check.

        Returns:
            MockColumnOperation for isnull.
        """
        if isinstance(column, str):
            column = MockColumn(column)
        return column.isnull()

    @staticmethod
    def isnotnull(column: Union[MockColumn, str]) -> MockColumnOperation:
        """Check if column value is not null.

        Args:
            column: Column to check.

        Returns:
            MockColumnOperation for isnotnull.
        """
        if isinstance(column, str):
            column = MockColumn(column)
        return column.isnotnull()

    @staticmethod
    def isnan(column: Union[MockColumn, str]) -> MockColumnOperation:
        """Check if column value is NaN.

        Args:
            column: Column to check.

        Returns:
            MockColumnOperation for isnan.
        """
        if isinstance(column, str):
            column = MockColumn(column)
        return MockColumnOperation(column, "isnan", None)

    @staticmethod
    def isnotnan(column: Union[MockColumn, str]) -> MockColumnOperation:
        """Check if column value is not NaN.

        Args:
            column: Column to check.

        Returns:
            MockColumnOperation for isnotnan.
        """
        if isinstance(column, str):
            column = MockColumn(column)
        return MockColumnOperation(column, "isnotnan", None)

    @staticmethod
    def expr(expr: str) -> MockColumnOperation:
        """Create a column expression from SQL string.

        Args:
            expr: SQL expression string.

        Returns:
            MockColumnOperation for the expression.
        """
        return MockColumnOperation(None, "expr", expr)

    @staticmethod
    def array(*columns: Union[MockColumn, str]) -> MockColumnOperation:
        """Create an array from columns.

        Args:
            *columns: Columns to include in array.

        Returns:
            MockColumnOperation for array.
        """
        col_refs = []
        for col in columns:
            if isinstance(col, str):
                col_refs.append(MockColumn(col))
            else:
                col_refs.append(col)

        return MockColumnOperation(None, "array", col_refs)

    @staticmethod
    def struct(*columns: Union[MockColumn, str]) -> MockColumnOperation:
        """Create a struct from columns.

        Args:
            *columns: Columns to include in struct.

        Returns:
            MockColumnOperation for struct.
        """
        col_refs = []
        for col in columns:
            if isinstance(col, str):
                col_refs.append(MockColumn(col))
            else:
                col_refs.append(col)

        return MockColumnOperation(None, "struct", col_refs)

    @staticmethod
    def greatest(*columns: Union[MockColumn, str]) -> MockColumnOperation:
        """Return the greatest value among columns.

        Args:
            *columns: Columns to compare.

        Returns:
            MockColumnOperation for greatest.
        """
        col_refs = []
        for col in columns:
            if isinstance(col, str):
                col_refs.append(MockColumn(col))
            else:
                col_refs.append(col)

        return MockColumnOperation(None, "greatest", col_refs)

    @staticmethod
    def least(*columns: Union[MockColumn, str]) -> MockColumnOperation:
        """Return the least value among columns.

        Args:
            *columns: Columns to compare.

        Returns:
            MockColumnOperation for least.
        """
        col_refs = []
        for col in columns:
            if isinstance(col, str):
                col_refs.append(MockColumn(col))
            else:
                col_refs.append(col)

        return MockColumnOperation(None, "least", col_refs)

    @staticmethod
    def when_otherwise(
        condition: MockColumnOperation, value: Any, otherwise: Any
    ) -> "MockCaseWhen":
        """Create a complete CASE WHEN expression.

        Args:
            condition: Condition to evaluate.
            value: Value if condition is true.
            otherwise: Default value.

        Returns:
            MockCaseWhen instance.
        """
        from ..conditional import MockCaseWhen

        case_when = MockCaseWhen(None, condition, value)
        return case_when.otherwise(otherwise)



"""
Storage manager module.

This module provides a unified storage manager that can use different backends.
"""

from typing import List, Dict, Any, Optional, Union
from datetime import datetime
import gc
import psutil
import os
from ..core.interfaces.storage import IStorageManager
from .backends.memory import MemoryStorageManager
from .backends.file import FileStorageManager
from mock_spark.backend.duckdb import DuckDBStorageManager
from mock_spark.spark_types import MockStructType, MockStructField


class TableMetadata:
    """Metadata for a table in the catalog."""

    def __init__(
        self,
        name: str,
        schema: str,
        created_at: datetime,
        table_schema: MockStructType,
        properties: Dict[str, Any],
    ):
        """Initialize table metadata.

        Args:
            name: Table name.
            schema: Schema/database name.
            created_at: Creation timestamp.
            table_schema: Table schema structure.
            properties: Table properties.
        """
        self.name = name
        self.schema = schema
        self.created_at = created_at
        self.table_schema = table_schema
        self.properties = properties
        self.qualified_name = f"{schema}.{name}"


class StorageManagerFactory:
    """Factory for creating storage managers."""

    @staticmethod
    def create_memory_manager() -> IStorageManager:
        """Create a memory storage manager.

        Returns:
            Memory storage manager instance.
        """
        return MemoryStorageManager()

    @staticmethod
    def create_file_manager(base_path: str = "mock_spark_storage") -> IStorageManager:
        """Create a file storage manager.

        Args:
            base_path: Base path for storage files.

        Returns:
            File storage manager instance.
        """
        return FileStorageManager(base_path)

    @staticmethod
    def create_duckdb_manager(db_path: Optional[str] = None) -> IStorageManager:
        """Create a DuckDB storage manager with in-memory storage by default.

        Args:
            db_path: Optional path to DuckDB database file. If None, uses in-memory storage.

        Returns:
            DuckDB storage manager instance.
        """
        return DuckDBStorageManager(db_path)


class UnifiedStorageManager(IStorageManager):
    """Unified storage manager that can switch between backends."""

    def __init__(self, backend: IStorageManager):
        """Initialize unified storage manager.

        Args:
            backend: Storage backend to use.
        """
        self.backend = backend
        self._table_metadata: Dict[str, TableMetadata] = {}

    def create_schema(self, schema: str) -> None:
        """Create a new schema.

        Args:
            schema: Name of the schema to create.
        """
        self.backend.create_schema(schema)

    def schema_exists(self, schema: str) -> bool:
        """Check if schema exists.

        Args:
            schema: Name of the schema to check.

        Returns:
            True if schema exists, False otherwise.
        """
        return self.backend.schema_exists(schema)

    def drop_schema(self, schema_name: str, cascade: bool = False) -> None:
        """Drop a schema.

        Args:
            schema_name: Name of the schema to drop.
            cascade: Whether to cascade the drop operation.
        """
        self.backend.drop_schema(schema_name, cascade)

    def list_schemas(self) -> List[str]:
        """List all schemas.

        Returns:
            List of schema names.
        """
        return self.backend.list_schemas()

    def table_exists(self, schema: str, table: str) -> bool:
        """Check if table exists.

        Args:
            schema: Name of the schema.
            table: Name of the table.

        Returns:
            True if table exists, False otherwise.
        """
        return self.backend.table_exists(schema, table)

    def create_table(
        self,
        schema: str,
        table: str,
        columns: Union[List[MockStructField], MockStructType],
    ) -> None:
        """Create a new table.

        Args:
            schema: Name of the schema.
            table: Name of the table.
            columns: Table columns definition.
        """
        self.backend.create_table(schema, table, columns)

    def drop_table(self, schema: str, table: str) -> None:
        """Drop a table.

        Args:
            schema: Name of the schema.
            table: Name of the table.
        """
        self.backend.drop_table(schema, table)

    def insert_data(
        self, schema: str, table: str, data: List[Dict[str, Any]], mode: str = "append"
    ) -> None:
        """Insert data into table.

        Args:
            schema: Name of the schema.
            table: Name of the table.
            data: Data to insert.
            mode: Insert mode ("append", "overwrite", "ignore").
        """
        self.backend.insert_data(schema, table, data)

    def query_table(
        self, schema: str, table: str, filter_expr: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """Query data from table.

        Args:
            schema: Name of the schema.
            table: Name of the table.
            filter_expr: Optional filter expression.

        Returns:
            List of data rows.
        """
        return self.backend.query_table(schema, table, filter_expr)

    def get_table_schema(
        self, schema_name: str, table_name: str
    ) -> Union[Any, MockStructType]:
        """Get table schema.

        Args:
            schema_name: Name of the schema.
            table_name: Name of the table.

        Returns:
            Table schema.
        """
        return self.backend.get_table_schema(schema_name, table_name)

    def get_data(self, schema: str, table: str) -> List[Dict[str, Any]]:
        """Get all data from table.

        Args:
            schema: Name of the schema.
            table: Name of the table.

        Returns:
            List of data rows.
        """
        return self.backend.get_data(schema, table)

    def create_temp_view(self, name: str, dataframe: Any) -> None:
        """Create a temporary view from a DataFrame.

        Args:
            name: Name of the temporary view.
            dataframe: DataFrame to create view from.
        """
        self.backend.create_temp_view(name, dataframe)

    def list_tables(self, schema_name: Optional[str] = None) -> List[str]:
        """List tables in schema.

        Args:
            schema_name: Name of the schema. If None, list tables in all schemas.

        Returns:
            List of table names.
        """
        return self.backend.list_tables(schema_name)

    def get_table_metadata(
        self, schema_name: str, table_name: str
    ) -> Union[Any, Dict[str, Any]]:
        """Get table metadata including Delta-specific fields.

        Args:
            schema_name: Name of the schema.
            table_name: Name of the table.

        Returns:
            Table metadata.
        """
        return self.backend.get_table_metadata(schema_name, table_name)

    def update_table_metadata(
        self, schema: str, table: str, metadata_updates: Dict[str, Any]
    ) -> None:
        """Update table metadata fields.

        Args:
            schema: Name of the schema.
            table: Name of the table.
            metadata_updates: Dictionary of metadata fields to update.
        """
        self.backend.update_table_metadata(schema, table, metadata_updates)

    def switch_backend(self, backend: IStorageManager) -> None:
        """Switch to a different storage backend.

        Args:
            backend: New storage backend to use.
        """
        self.backend = backend

    def save_table_metadata(self, qualified_name: str, metadata: TableMetadata) -> None:
        """Store table metadata.

        Args:
            qualified_name: Qualified table name (schema.table).
            metadata: Table metadata to store.
        """
        self._table_metadata[qualified_name] = metadata

    def get_table_metadata_by_name(
        self, qualified_name: str
    ) -> Optional[TableMetadata]:
        """Retrieve table metadata by qualified name.

        Args:
            qualified_name: Qualified table name (schema.table).

        Returns:
            Table metadata or None if not found.
        """
        return self._table_metadata.get(qualified_name)

    def list_table_metadata(self, schema: str) -> List[TableMetadata]:
        """List all table metadata for a schema.

        Args:
            schema: Schema name.

        Returns:
            List of table metadata objects.
        """
        return [
            metadata
            for qualified_name, metadata in self._table_metadata.items()
            if metadata.schema == schema
        ]

    def cleanup_temp_tables(self) -> None:
        """Clean up temporary tables to free memory."""
        try:
            if hasattr(self.backend, "cleanup_temp_tables"):
                self.backend.cleanup_temp_tables()
        except Exception:
            # Ignore errors during cleanup
            pass

    def optimize_storage(self) -> None:
        """Optimize storage by cleaning up and compacting data."""
        try:
            if hasattr(self.backend, "optimize_storage"):
                self.backend.optimize_storage()
        except Exception:
            # Ignore errors during optimization
            pass

    def get_memory_usage(self) -> Dict[str, Any]:
        """Get current memory usage statistics.

        Returns:
            Dictionary with memory usage information.
        """
        try:
            process = psutil.Process(os.getpid())
            memory_info = process.memory_info()

            return {
                "rss": memory_info.rss,  # Resident Set Size
                "vms": memory_info.vms,  # Virtual Memory Size
                "percent": process.memory_percent(),
                "available": psutil.virtual_memory().available,
                "total": psutil.virtual_memory().total,
            }
        except Exception:
            return {"rss": 0, "vms": 0, "percent": 0, "available": 0, "total": 0}

    def force_garbage_collection(self) -> None:
        """Force garbage collection to free memory."""
        gc.collect()

    def get_table_sizes(self) -> Dict[str, int]:
        """Get estimated sizes of all tables.

        Returns:
            Dictionary mapping table names to estimated sizes.
        """
        sizes = {}
        for qualified_name, metadata in self._table_metadata.items():
            try:
                # Estimate size based on metadata
                estimated_size = (
                    len(metadata.table_schema.fields) * 100
                )  # Rough estimate
                sizes[qualified_name] = estimated_size
            except Exception:
                sizes[qualified_name] = 0
        return sizes

    def cleanup_old_tables(self, max_age_hours: int = 24) -> int:
        """Clean up tables older than specified age.

        Args:
            max_age_hours: Maximum age in hours before cleanup.

        Returns:
            Number of tables cleaned up.
        """
        from datetime import timedelta

        cutoff_time = datetime.now() - timedelta(hours=max_age_hours)
        cleaned_count = 0

        for qualified_name, metadata in list(self._table_metadata.items()):
            if metadata.created_at < cutoff_time:
                try:
                    # Remove from metadata
                    del self._table_metadata[qualified_name]

                    # Remove from backend if possible
                    if hasattr(self.backend, "drop_table"):
                        schema, table = qualified_name.split(".", 1)
                        self.backend.drop_table(schema, table)

                    cleaned_count += 1
                except Exception:
                    # Ignore errors during cleanup
                    pass

        return cleaned_count



"""
PySpark version compatibility module.

This module controls which functions and methods are available based on
the PySpark version compatibility mode set during installation.

By default, all features are available. When installed with a version specifier
like `pip install mock-spark[pyspark-3-1]`, only APIs available in that
PySpark version will be exposed.
"""

import json
import os
from pathlib import Path
from typing import Dict, Optional


# Global version setting - None means all features available
_PYSPARK_COMPAT_VERSION: Optional[str] = None

# Cache for API matrix
_API_MATRIX: Optional[Dict[str, Dict[str, Dict[str, bool]]]] = None


def set_pyspark_version(version: Optional[str]) -> None:
    """
    Set PySpark compatibility version.

    Args:
        version: PySpark version string (e.g., '3.1', '3.2', '3.5') or None for all features

    Examples:
        >>> set_pyspark_version('3.1')
        >>> # Only PySpark 3.1-compatible APIs will be available
    """
    global _PYSPARK_COMPAT_VERSION

    # Handle None - reset to all features
    if version is None:
        _PYSPARK_COMPAT_VERSION = None
        return

    # Normalize version format (3.1 or 3.1.3 -> 3.1.3)
    if version.count(".") == 1:
        # Map major.minor to specific patch version
        version_map = {
            "3.0": "3.0.3",
            "3.1": "3.1.3",
            "3.2": "3.2.4",
            "3.3": "3.3.4",
            "3.4": "3.4.3",
            "3.5": "3.5.2",
        }
        version = version_map.get(version, version)

    _PYSPARK_COMPAT_VERSION = version


def get_pyspark_version() -> Optional[str]:
    """
    Get current PySpark compatibility version.

    Returns:
        Version string or None if all features available
    """
    return _PYSPARK_COMPAT_VERSION


def _load_api_matrix() -> Dict[str, Dict[str, Dict[str, bool]]]:
    """
    Load the API availability matrix from JSON file.

    Returns:
        Dictionary with 'functions' and 'dataframe_methods' matrices
    """
    global _API_MATRIX

    if _API_MATRIX is not None:
        return _API_MATRIX

    # Load from package data
    matrix_path = Path(__file__).parent / "pyspark_api_matrix.json"

    if not matrix_path.exists():
        # Fallback: all items available if matrix file missing
        return {"functions": {}, "dataframe_methods": {}}

    with open(matrix_path, "r") as f:
        _API_MATRIX = json.load(f)

    return _API_MATRIX


def is_available(item_name: str, item_type: str = "function") -> bool:
    """
    Check if a function or method is available in current compatibility mode.

    Args:
        item_name: Name of the function or method
        item_type: Either 'function' or 'dataframe_method'

    Returns:
        True if item is available in current version mode

    Examples:
        >>> set_pyspark_version('3.0')
        >>> is_available('make_date', 'function')
        False
        >>> is_available('overlay', 'function')
        True
    """
    # If no version set, all features available
    if _PYSPARK_COMPAT_VERSION is None:
        return True

    # Load matrix
    matrix = _load_api_matrix()

    # Determine which matrix to check
    if item_type == "function":
        item_matrix = matrix.get("functions", {})
    elif item_type == "dataframe_method":
        item_matrix = matrix.get("dataframe_methods", {})
    else:
        # Unknown type, allow by default
        return True

    # Check if item exists in matrix
    if item_name not in item_matrix:
        # Item not in matrix, allow by default (might be mock-spark specific)
        return True

    # Check if available in target version
    availability = item_matrix[item_name]
    return availability.get(_PYSPARK_COMPAT_VERSION, False)


def check_version_from_env() -> None:
    """
    Check environment variable for version setting.

    Looks for MOCK_SPARK_PYSPARK_VERSION environment variable.
    """
    env_version = os.environ.get("MOCK_SPARK_PYSPARK_VERSION")
    if env_version:
        set_pyspark_version(env_version)


def check_version_from_marker() -> None:
    """
    Check for version marker file in package directory.

    Looks for .pyspark_X_Y_compat marker files created during installation.
    """
    marker_dir = Path(__file__).parent

    # Check for marker files in order (latest first)
    version_markers = [
        (".pyspark_3_5_compat", "3.5"),
        (".pyspark_3_4_compat", "3.4"),
        (".pyspark_3_3_compat", "3.3"),
        (".pyspark_3_2_compat", "3.2"),
        (".pyspark_3_1_compat", "3.1"),
        (".pyspark_3_0_compat", "3.0"),
    ]

    for marker_file, version in version_markers:
        if (marker_dir / marker_file).exists():
            set_pyspark_version(version)
            return


# Auto-check version on module import
check_version_from_env()
check_version_from_marker()



"""
Table management utilities for Mock Spark.

This module provides table lifecycle management for DuckDB operations.
"""

from typing import Any, Dict, List, Optional
from sqlalchemy import (
    Column,
    Integer,
    Float,
    Double,
    Boolean,
    String,
    DateTime,
    ARRAY,
    VARCHAR,
    Table,
    text,
)
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from sqlalchemy.sql.schema import MetaData as MetaDataType

from ...spark_types import MockRow


class DuckDBTableManager:
    """Manages table lifecycle for DuckDB operations."""

    def __init__(self, engine: Engine, metadata: MetaDataType):
        """Initialize table manager.

        Args:
            engine: SQLAlchemy engine instance
            metadata: SQLAlchemy metadata instance
        """
        self.engine = engine
        self.metadata = metadata
        self._created_tables: Dict[str, Any] = {}

    def create_table_with_data(
        self, table_name: str, data: List[Dict[str, Any]]
    ) -> Table:
        """Create a table and insert data using SQLAlchemy Table.

        Args:
            table_name: Name of the table to create
            data: Data to insert into the table

        Returns:
            Created SQLAlchemy Table object
        """
        if not data:
            # Create a minimal table with at least one column to avoid "Table must have at least one column!" error
            columns = [Column("id", Integer)]
            table = Table(table_name, self.metadata, *columns)
            table.create(self.engine, checkfirst=True)
            self._created_tables[table_name] = table
            return table

        # Create table using SQLAlchemy Table approach
        columns = []
        has_map_columns = False
        map_column_names = []

        if data:
            for key, value in data[0].items():
                if isinstance(value, int):
                    columns.append(Column(key, Integer))
                elif isinstance(value, float):
                    columns.append(Column(key, Float))  # type: ignore[arg-type]
                elif isinstance(value, bool):
                    columns.append(Column(key, Boolean))  # type: ignore[arg-type]
                elif isinstance(value, list):
                    # For arrays, infer element type from first element
                    if value and len(value) > 0:
                        # Infer element type from first non-None element
                        first_elem = next(
                            (elem for elem in value if elem is not None), None
                        )
                        if isinstance(first_elem, int):
                            columns.append(Column(key, ARRAY(Integer)))
                        elif isinstance(first_elem, float):
                            columns.append(Column(key, ARRAY(Float)))
                        elif isinstance(first_elem, bool):
                            columns.append(Column(key, ARRAY(Boolean)))
                        else:
                            columns.append(Column(key, ARRAY(VARCHAR)))
                    else:
                        # Empty array - default to VARCHAR
                        columns.append(Column(key, ARRAY(VARCHAR)))
                elif isinstance(value, dict):
                    # For maps, mark for raw SQL handling
                    has_map_columns = True
                    map_column_names.append(key)
                    columns.append(Column(key, String))  # type: ignore[arg-type] # Placeholder
                else:
                    columns.append(Column(key, String))  # type: ignore[arg-type]

        # Create table - use raw SQL for MAP/ARRAY columns
        has_array_columns = any(
            type(col.type).__name__ == "ArrayType" for col in columns
        )
        if has_map_columns or has_array_columns:
            # Build CREATE TABLE with proper MAP/ARRAY types using raw SQL
            col_defs = []
            for col in columns:
                if col.name in map_column_names:
                    col_defs.append(f'"{col.name}" MAP(VARCHAR, VARCHAR)')
                elif type(col.type).__name__ == "ArrayType":
                    # Determine array element type
                    if hasattr(col.type, "element_type"):
                        elem_type = col.type.element_type
                    else:
                        # Fallback for non-array types
                        col_defs.append(f'"{col.name}" VARCHAR[]')
                        continue
                    if elem_type.__class__.__name__ == "LongType":
                        col_defs.append(f'"{col.name}" INTEGER[]')
                    elif elem_type.__class__.__name__ in ["FloatType", "DoubleType"]:
                        col_defs.append(f'"{col.name}" DOUBLE[]')
                    elif elem_type.__class__.__name__ == "BooleanType":
                        col_defs.append(f'"{col.name}" BOOLEAN[]')
                    else:
                        col_defs.append(f'"{col.name}" VARCHAR[]')
                elif isinstance(col.type, Integer):
                    col_defs.append(f'"{col.name}" INTEGER')
                elif isinstance(col.type, Float) or isinstance(col.type, Double):
                    col_defs.append(f'"{col.name}" DOUBLE')
                elif isinstance(col.type, Boolean):
                    col_defs.append(f'"{col.name}" BOOLEAN')
                else:
                    col_defs.append(f'"{col.name}" VARCHAR')

            create_sql = f"CREATE TABLE {table_name} ({', '.join(col_defs)})"
            with Session(self.engine) as session:
                session.execute(text(create_sql))
                session.commit()

            # Register table in metadata manually
            table = Table(table_name, self.metadata, *columns, extend_existing=True)
            self._created_tables[table_name] = table
        else:
            # Normal table creation
            table = Table(table_name, self.metadata, *columns)
            table.create(self.engine, checkfirst=True)
            self._created_tables[table_name] = table

        # Insert data using raw SQL - handle dict/list conversion for DuckDB
        with Session(self.engine) as session:
            for row_data in data:
                # Convert row data to values for insert, handling special types
                insert_values: Dict[str, Any] = {}
                for col in columns:
                    value = row_data[col.name]

                    # Convert Python dict to DuckDB MAP
                    if isinstance(value, dict):
                        # Convert dict to MAP syntax: MAP(['keys'], ['values'])
                        if value:
                            keys = list(value.keys())
                            vals = list(value.values())
                            # Create MAP using raw SQL
                            map_sql = f"MAP({keys!r}, {vals!r})"
                            insert_values[col.name] = text(map_sql)
                        else:
                            insert_values[col.name] = None
                    # Convert Python list to DuckDB ARRAY
                    elif isinstance(value, list):
                        # Convert list to ARRAY syntax: [element1, element2, ...]
                        if value:
                            # Escape string values properly
                            array_elements = []
                            for elem in value:
                                if isinstance(elem, str):
                                    array_elements.append(f"'{elem}'")
                                else:
                                    array_elements.append(str(elem))
                            array_sql = f"[{', '.join(array_elements)}]"
                            insert_values[col.name] = text(array_sql)
                        else:
                            insert_values[col.name] = None
                    else:
                        insert_values[col.name] = value

                # Insert using parameterized values for non-MAP/ARRAY columns
                # and raw SQL for MAP/ARRAY columns
                if any(isinstance(v, type(text(""))) for v in insert_values.values()):
                    # Has MAP columns - use raw SQL
                    col_names = []
                    col_values = []
                    for col_name, col_value in insert_values.items():
                        col_names.append(f'"{col_name}"')
                        if hasattr(col_value, "text"):  # TextClause
                            col_values.append(col_value.text)
                        elif isinstance(col_value, str):
                            col_values.append(f"'{col_value}'")
                        elif col_value is None:
                            col_values.append("NULL")
                        else:
                            col_values.append(str(col_value))

                    raw_sql = f"INSERT INTO {table_name} ({', '.join(col_names)}) VALUES ({', '.join(col_values)})"
                    session.execute(text(raw_sql))
                else:
                    # Normal insert - handle overflow values gracefully
                    try:
                        insert_stmt = table.insert().values(insert_values)
                        session.execute(insert_stmt)
                    except Exception as e:
                        if "out of range" in str(e) or "Conversion Error" in str(e):
                            # Rollback the failed transaction
                            session.rollback()

                            # Handle overflow by inserting NULL for problematic values
                            safe_values: Dict[str, Any] = {}
                            for key, value in insert_values.items():
                                col_type = table.c[key].type
                                # Check if this is an Integer column and value is too large
                                if isinstance(col_type, Integer):
                                    if isinstance(value, (int, float)):
                                        # Check if value exceeds INT32 range
                                        if value > 2147483647 or value < -2147483648:
                                            safe_values[key] = None
                                        else:
                                            safe_values[key] = value
                                    else:
                                        safe_values[key] = value
                                else:
                                    safe_values[key] = value

                            if safe_values:
                                safe_insert_stmt = table.insert().values(safe_values)
                                session.execute(safe_insert_stmt)
                        else:
                            raise e
            session.commit()

        return table

    def copy_table_structure(self, source_table: str, target_table: str) -> None:
        """Copy table structure from source to target.

        Args:
            source_table: Name of source table
            target_table: Name of target table
        """
        source_table_obj = self._created_tables[source_table]

        # Copy all columns from source table
        new_columns: List[Any] = []
        for column in source_table_obj.columns:
            new_columns.append(Column(column.name, column.type, primary_key=False))

        # Create target table using SQLAlchemy Table
        target_table_obj = Table(target_table, self.metadata, *new_columns)
        target_table_obj.create(self.engine, checkfirst=True)
        self._created_tables[target_table] = target_table_obj

    def get_table_results(self, table_name: str) -> List[MockRow]:
        """Get all results from a table as MockRow objects.

        Args:
            table_name: Name of the table to query

        Returns:
            List of MockRow objects
        """
        table_obj = self._created_tables[table_name]

        with Session(self.engine) as session:
            # Build raw SQL query
            # Escape double quotes in column names by doubling them
            column_names = [
                f'"{col.name.replace(chr(34), chr(34) + chr(34))}"'
                for col in table_obj.columns
            ]
            sql = f"SELECT {', '.join(column_names)} FROM {table_name}"
            results = session.execute(text(sql)).all()

            mock_rows = []
            for result in results:
                # Convert result to list of values in schema order (preserves duplicates)
                # Don't use dict to avoid losing duplicate column names
                row_values: List[Any] = []
                field_names = []

                for i, column in enumerate(table_obj.columns):
                    value = result[i]

                    # Convert value to appropriate type based on column type
                    is_array_column = (
                        isinstance(column.type, ARRAY)
                        or type(column.type).__name__ == "ARRAY"
                    )
                    if is_array_column and value is not None:
                        # Array columns might be returned as lists or strings
                        if isinstance(value, list):
                            row_values.append(value)
                        elif isinstance(value, str):
                            # Parse string representation back to list
                            # DuckDB sometimes returns arrays as strings like "[1, 2, 3]"
                            try:
                                import ast

                                row_values.append(ast.literal_eval(value))
                            except (ValueError, SyntaxError):
                                row_values.append(value)
                        else:
                            row_values.append(value)
                    elif (
                        isinstance(column.type, String)
                        and isinstance(value, str)
                        and value.startswith("{")
                        and value.endswith("}")
                    ):
                        # Map columns returned as strings like "{a=1, b=2}" - parse to dict
                        try:
                            import ast

                            # Try to parse as dict literal
                            row_values.append(ast.literal_eval(value))
                        except (ValueError, SyntaxError):
                            # If that fails, leave as string
                            row_values.append(value)
                    elif isinstance(value, dict):
                        # Already a dict (map)
                        row_values.append(value)
                    elif isinstance(column.type, Integer) and value is not None:
                        try:
                            row_values.append(int(value))
                        except (ValueError, TypeError):
                            row_values.append(value)
                    elif isinstance(column.type, Float) and value is not None:
                        try:
                            row_values.append(float(value))
                        except (ValueError, TypeError):
                            row_values.append(value)
                    elif isinstance(column.type, Boolean) and value is not None:
                        if isinstance(value, str):
                            row_values.append(
                                value.lower()
                                in (
                                    "true",
                                    "1",
                                    "yes",
                                    "on",
                                )
                            )
                        else:
                            row_values.append(bool(value))
                    elif isinstance(column.type, DateTime) and value is not None:
                        # Preserve datetime/timestamp types - don't convert to string
                        row_values.append(value)
                    else:
                        row_values.append(value)

                    field_names.append(column.name)

                # Create MockRow with values in order
                # Convert to list of tuples to preserve duplicate column names
                row_data = list(zip(field_names, row_values))
                mock_rows.append(MockRow(row_data))

            return mock_rows

    def get_table(self, table_name: str) -> Optional[Table]:
        """Get a table object by name.

        Args:
            table_name: Name of the table

        Returns:
            SQLAlchemy Table object or None if not found
        """
        return self._created_tables.get(table_name)

    def cleanup_temp_tables(self) -> None:
        """Clean up all temporary tables."""
        with Session(self.engine) as session:
            for table_name in list(self._created_tables.keys()):
                try:
                    session.execute(text(f"DROP TABLE IF EXISTS {table_name}"))
                    session.commit()
                except Exception:
                    # Ignore errors during cleanup
                    pass
        self._created_tables.clear()



"""Set operations for MockDataFrame."""

from typing import List, Any, Dict, Tuple, TYPE_CHECKING

from ...spark_types import MockRow

if TYPE_CHECKING:
    from ...spark_types import MockStructType


class SetOperations:
    """Handles set operations for MockDataFrame."""

    @staticmethod
    def distinct_rows(rows: List[MockRow]) -> List[MockRow]:
        """Remove duplicate rows from a list."""
        seen = set()
        unique_rows = []
        for row in rows:
            # Create a hashable representation of the row
            row_values = []
            for col in row.__dict__.keys():
                value = getattr(row, col)
                # Convert unhashable types to hashable representations
                if isinstance(value, dict):
                    value = tuple(sorted(value.items()))
                elif isinstance(value, list):
                    value = tuple(value)
                row_values.append(value)
            row_tuple = tuple(row_values)
            if row_tuple not in seen:
                seen.add(row_tuple)
                unique_rows.append(row)
        return unique_rows

    @staticmethod
    def union_rows(rows1: List[MockRow], rows2: List[MockRow]) -> List[MockRow]:
        """Union two lists of rows."""
        return rows1 + rows2

    @staticmethod
    def union(
        data1: List[Dict[str, Any]],
        schema1: "MockStructType",
        data2: List[Dict[str, Any]],
        schema2: "MockStructType",
        storage: Any,
    ) -> Tuple[List[Dict[str, Any]], "MockStructType"]:
        """Union two DataFrames with their data and schemas."""
        # Convert data to MockRow objects for union
        rows1 = [MockRow(row) for row in data1]
        rows2 = [MockRow(row) for row in data2]

        # Perform union
        unioned_rows = SetOperations.union_rows(rows1, rows2)

        # Convert back to dict format
        result_data = [row.__dict__ for row in unioned_rows]

        # Return the schema from the first DataFrame (assuming they're compatible)
        return result_data, schema1

    @staticmethod
    def intersect_rows(rows1: List[MockRow], rows2: List[MockRow]) -> List[MockRow]:
        """Find intersection of two lists of rows."""

        def make_hashable(row: MockRow) -> Tuple[Any, ...]:
            row_values = []
            for col in row.__dict__.keys():
                value = getattr(row, col)
                if isinstance(value, dict):
                    value = tuple(sorted(value.items()))
                elif isinstance(value, list):
                    value = tuple(value)
                row_values.append(value)
            return tuple(row_values)

        set1 = {make_hashable(row) for row in rows1}
        set2 = {make_hashable(row) for row in rows2}

        intersection = set1.intersection(set2)
        result = []
        for row in rows1:
            row_tuple = make_hashable(row)
            if row_tuple in intersection:
                result.append(row)
        return result

    @staticmethod
    def except_rows(rows1: List[MockRow], rows2: List[MockRow]) -> List[MockRow]:
        """Find rows in rows1 that are not in rows2."""

        def make_hashable(row: MockRow) -> Tuple[Any, ...]:
            row_values = []
            for col in row.__dict__.keys():
                value = getattr(row, col)
                if isinstance(value, dict):
                    value = tuple(sorted(value.items()))
                elif isinstance(value, list):
                    value = tuple(value)
                row_values.append(value)
            return tuple(row_values)

        set2 = {make_hashable(row) for row in rows2}
        result = []
        for row in rows1:
            row_tuple = make_hashable(row)
            if row_tuple not in set2:
                result.append(row)
        return result

    @staticmethod
    def rows_equal(row1: MockRow, row2: MockRow) -> bool:
        """Check if two rows are equal."""
        if row1.__dict__.keys() != row2.__dict__.keys():
            return False
        return all(
            getattr(row1, col) == getattr(row2, col) for col in row1.__dict__.keys()
        )



"""DataFrame formatting utilities for display operations."""

from typing import Any, List, Optional

from ...spark_types import MockRow


class DataFrameFormatter:
    """Handles DataFrame formatting for display operations."""

    @staticmethod
    def format_row(row: MockRow, columns: List[str]) -> str:
        """Format a single row for display."""
        formatted_values = []
        for col in columns:
            value = getattr(row, col, None)
            if value is None:
                formatted_values.append("null")
            elif isinstance(value, str):
                formatted_values.append(f'"{value}"')
            else:
                formatted_values.append(str(value))
        return f"[{', '.join(formatted_values)}]"

    @staticmethod
    def format_rows(
        rows: List[MockRow], columns: List[str], limit: Optional[int] = None
    ) -> List[str]:
        """Format multiple rows for display."""
        if limit is not None:
            rows = rows[:limit]
        return [DataFrameFormatter.format_row(row, columns) for row in rows]

    @staticmethod
    def format_schema(columns: List[str], types: List[str]) -> str:
        """Format the schema information."""
        schema_parts = []
        for col, col_type in zip(columns, types):
            schema_parts.append(f" {col}: {col_type}")
        return "root\n" + "\n".join(schema_parts)

    @staticmethod
    def truncate_string(value: str, max_length: int = 50) -> str:
        """Truncate a string to the specified maximum length."""
        if len(value) <= max_length:
            return value
        return value[: max_length - 3] + "..."

    @staticmethod
    def format_value_for_display(value: Any, max_length: int = 50) -> str:
        """Format a value for display with truncation."""
        if value is None:
            return "null"
        elif isinstance(value, str):
            return DataFrameFormatter.truncate_string(value, max_length)
        else:
            return str(value)
